# ImageNet Emoji Dictionary for the Logit Lens

IMAGENET_EMOJI = {
    0: '🐟',  # tench
    1: '🐠',  # goldfish
    2: '🦈',  # great white shark
    3: '🦈',  # tiger shark
    4: '🦈',  # hammerhead shark
    5: '🐡',  # electric ray
    6: '🐠',  # stingray
    7: '🐓',  # cock
    8: '🐔',  # hen
    9: '🐦',  # ostrich
    10: '🐦',  # brambling
    11: '🐦',  # goldfinch
    12: '🐦',  # house finch
    13: '🐦',  # junco
    14: '🐦',  # indigo bunting
    15: '🐦',  # robin
    16: '🐦',  # bulbul
    17: '🐦',  # jay
    18: '🐦',  # magpie
    19: '🐦',  # chickadee
    20: '🐦',  # water ouzel
    21: '🪁',  # kite
    22: '🦅',  # bald eagle
    23: '🦅',  # vulture
    24: '🦉',  # great grey owl
    25: '🦎',  # European fire salamander
    26: '🦎',  # common newt
    27: '🦎',  # eft
    28: '🦎',  # spotted salamander
    29: '🦎',  # axolotl
    30: '🐸',  # bullfrog
    31: '🐸',  # tree frog
    32: '🐸',  # tailed frog
    33: '🐢',  # loggerhead turtle
    34: '🐢',  # leatherback turtle
    35: '🐢',  # mud turtle
    36: '🐢',  # terrapin
    37: '🐢',  # box turtle
    38: '🦎',  # banded gecko
    39: '🦎',  # common iguana
    40: '🦎',  # American chameleon
    41: '🦎',  # whiptail
    42: '🦎',  # agama
    43: '🦎',  # frilled lizard
    44: '🦎',  # alligator lizard
    45: '🦎',  # Gila monster
    46: '🦎',  # green lizard
    47: '🦎',  # African chameleon
    48: '🦎',  # Komodo dragon
    49: '🐊',  # African crocodile
    50: '🐊',  # American alligator
    51: '🦕',  # triceratops
    52: '🐍',  # thunder snake
    53: '🐍',  # ringneck snake
    54: '🐍',  # hognose snake
    55: '🐍',  # green snake
    56: '🐍',  # king snake
    57: '🐍',  # garter snake
    58: '🐍',  # water snake
    59: '🐍',  # vine snake
    60: '🐍',  # night snake
    61: '🐍',  # boa constrictor
    62: '🐍',  # rock python
    63: '🐍',  # Indian cobra
    64: '🐍',  # green mamba
    65: '🐍',  # sea snake
    66: '🐍',  # horned viper
    67: "🐍",  # diamondback, diamondback rattlesnake
    68: "🐍",  # sidewinder, horned rattlesnake
    69: "🦖",  # trilobite, using a dinosaur emoji as a proxy for ancient creatures
    70: "🕷️",  # harvestman, daddy longlegs
    71: "🦂",  # scorpion
    72: "🕷️",  # black and gold garden spider
    73: "🕷️",  # barn spider
    74: "🕷️",  # garden spider
    75: "🕷️",  # black widow
    76: "🕷️",  # tarantula
    77: "🕷️",  # wolf spider, hunting spider
    78: "🕷️",  # tick, using spider emoji as a proxy
    79: "🐛",  # centipede
    80: "🐦",  # black grouse
    81: "🐦",  # ptarmigan
    82: "🐦",  # ruffed grouse, partridge
    83: "🐦",  # prairie chicken, prairie grouse, prairie fowl
    84: "🦚",  # peacock
    85: "🐦",  # quail
    86: "🐦",  # partridge
    87: "🦜",  # African grey, African gray
    88: "🦜",  # macaw
    89: "🦜",  # sulphur-crested cockatoo
    90: "🦜",  # lorikeet
    91: "🐦",  # coucal
    92: "🐦",  # bee eater
    93: "🐦",  # hornbill
    94: "🐦",  # hummingbird
    95: "🐦",  # jacamar
    96: "🐦",  # toucan
    97: "🦆",  # drake
    98: "🦆",  # red-breasted merganser
    99: "🦆",  # goose
    100: "🦢",  # black swan
    101: "🐘",  # tusker, using elephant emoji as a proxy
    102: "🦔",  # echidna, spiny anteater, using hedgehog emoji as a proxy
    103: "🦆",  # platypus, using duck emoji as a proxy
    104: "🦘",  # wallaby, brush kangaroo
    105: "🐨",  # koala
    106: "🦡",  # wombat, using badger emoji as a proxy
    107: "🦑",  # jellyfish, using squid emoji as a proxy
    108: "🐚",  # sea anemone
    109: "🪸",  # brain coral
    110: "🪱",  # flatworm, platyhelminth
    111: "🪱",  # nematode, nematode worm, roundworm
    112: "🐚",  # conch
    113: "🐌",  # snail
    114: "🐌",  # slug, using snail emoji as a proxy
    115: "🐌",  # sea slug, nudibranch, using snail emoji as a proxy
    116: "🐚",  # chiton, coat-of-mail shell, sea cradle, polyplacophore
    117: "🐚",  # chambered nautilus, pearly nautilus
    118: "🦀",  # Dungeness crab
    119: "🦀",  # rock crab
    120: "🦀",  # fiddler crab
    121: "🦀",  # king crab, Alaska crab
    122: "🦞",  # American lobster
    123: "🦞",  # spiny lobster,
    124: "🦞",  # crayfish, crawfish, crawdad, crawdaddy
    125: "🦀",  # hermit crab
    126: "🦐",  # isopod
    127: "🐦",  # white stork, Ciconia ciconia
    128: "🐦",  # black stork, Ciconia nigra
    129: "🐦",  # spoonbill
    130: "🦩",  # flamingo
    131: "🐦",  # little blue heron, Egretta caerulea
    132: "🐦",  # American egret, great white heron, Egretta albus
    133: "🐦",  # bittern
    134: "🐦",  # crane
    135: "🐦",  # limpkin, Aramus pictus
    136: "🐦",  # European gallinule, Porphyrio porphyrio
    137: "🐦",  # American coot, marsh hen, mud hen, water hen, Fulica americana
    138: "🐦",  # bustard
    139: "🐦",  # ruddy turnstone, Arenaria interpres
    140: "🐦",  # red-backed sandpiper, dunlin, Erolia alpina
    141: "🐦",  # redshank, Tringa totanus
    142: "🐦",  # dowitcher
    143: "🐦",  # oystercatcher, oyster catcher
    144: "🐦",  # pelican
    145: "🐧",  # king penguin, Aptenodytes patagonica
    146: "🐦",  # albatross, mollymawk
    147: "🐋",  # grey whale, gray whale, devilfish, Eschrichtius gibbosus, Eschrichtius robustus
    148: "🐋",  # killer whale, killer, orca, grampus, sea wolf, Orcinus orca
    149: "🐋",  # dugong, Dugong dugon
    150: "🦭",  # sea lion
    151: "🐶",  # Chihuahua
    152: "🐶",  # Japanese spaniel
    153: "🐶",  # Maltese dog, Maltese terrier, Maltese
    154: "🐶",  # Pekinese, Pekingese, Peke
    155: "🐶",  # Shih-Tzu
    156: "🐶",  # Blenheim spaniel
    157: "🐶",  # papillon
    158: "🐶",  # toy terrier
    159: "🐶",  # Rhodesian ridgeback
    160: "🐶",  # Afghan hound, Afghan
    161: "🐶",  # basset, basset hound
    162: "🐶",  # beagle
    163: "🐶",  # bloodhound, sleuthhound
    164: "🐶",  # bluetick
    165: "🐶",  # black-and-tan coonhound
    166: "🐶",  # Walker hound, Walker foxhound
    167: "🐶",  # English foxhound
    168: "🐶",  # redbone
    169: "🐶",  # borzoi, Russian wolfhound
    170: "🐶",  # Irish wolfhound
    171: "🐶",  # Italian greyhound
    172: "🐶",  # whippet
    173: "🐶",  # Ibizan hound, Ibizan Podenco
    174: "🐶",  # Norwegian elkhound, elkhound
    175: "🐶",  # otterhound, otter hound
    176: "🐶",  # Saluki, gazelle hound
    177: "🐶",  # Scottish deerhound, deerhound
    178: "🐶",  # Weimaraner
    179: "🐶",  # Staffordshire bullterrier, Staffordshire bull terrier
    180: "🐶",  # American Staffordshire terrier, Staffordshire terrier, American pit bull terrier, pit bull terrier
    181: "🐶",  # Bedlington terrier
    182: "🐶",  # Border terrier
    183: "🐶",  # Kerry blue terrier
    184: "🐶",  # Irish terrier
    185: "🐶",  # Norfolk terrier
    186: "🐶",  # Norwich terrier
    187: "🐶",  # Yorkshire terrier
    188: "🐶",  # wire-haired fox terrier
    189: "🐶",  # Lakeland terrier
    190: "🐶",  # Sealyham terrier, Sealyham
    191: "🐶",  # Airedale, Airedale terrier
    192: "🐶",  # cairn, cairn terrier
    193: "🐶",  # Australian terrier
    194: "🐶",  # Dandie Dinmont, Dandie Dinmont terrier
    195: "🐶",  # Boston bull, Boston terrier
    196: "🐶",  # miniature schnauzer
    197: "🐶",  # giant schnauzer
    198: "🐶",  # standard schnauzer
    199: "🐶",  # Scotch terrier, Scottish terrier, Scottie
    200: "🐶",  # Tibetan terrier, chrysanthemum dog
    201: "🐶",  # silky terrier, Sydney silky
    202: "🐶",  # soft-coated wheaten terrier
    203: "🐶",  # West Highland white terrier
    204: "🐶",  # Lhasa, Lhasa apso
    205: "🐶",  # flat-coated retriever
    206: "🐶",  # curly-coated retriever
    207: "🐶",  # golden retriever
    208: "🐶",  # Labrador retriever
    209: "🐶",  # Chesapeake Bay retriever
    210: "🐶",  # German short-haired pointer
    211: "🐶",  # vizsla, Hungarian pointer
    212: "🐶",  # English setter
    213: "🐶",  # Irish setter, red setter
    214: "🐶",  # Gordon setter
    215: "🐶",  # Brittany spaniel
    216: "🐶",  # clumber, clumber spaniel
    217: "🐶",  # English springer, English springer spaniel
    218: "🐶",  # Welsh springer spaniel
    219: "🐶",  # cocker spaniel, English cocker spaniel, cocker
    220: "🐶",  # Sussex spaniel
    221: "🐶",  # Irish water spaniel
    222: "🐶",  # kuvasz
    223: "🐶",  # schipperke
    224: "🐶",  # groenendael
    225: "🐶",  # malinois
    226: "🐶",  # briard
    227: "🐶",  # kelpie
    228: "🐶",  # komondor
    229: "🐶",  # Old English sheepdog, bobtail
    230: "🐶",  # Shetland sheepdog, Shetland sheep dog, Shetland
    231: "🐶",  # collie
    232: "🐶",  # Border collie
    233: "🐶",  # Bouvier des Flandres, Bouviers des Flandres
    234: "🐶",  # Rottweiler
    235: "🐶",  # German shepherd, German shepherd dog, German police dog, alsatian
    236: "🐶",  # Doberman, Doberman pinscher
    237: "🐶",  # miniature pinscher
    238: "🐶",  # Greater Swiss Mountain dog
    239: "🐶",  # Bernese mountain dog
    240: "🐶",  # Appenzeller
    241: "🐶",  # EntleBucher
    242: "🐶",  # boxer
    243: "🐶",  # bull mastiff
    244: "🐶",  # Tibetan mastiff
    245: "🐶",  # French bulldog
    246: "🐶",  # Great Dane
    247: "🐶",  # Saint Bernard, St Bernard
    248: "🐶",  # Eskimo dog, husky
    249: "🐶",  # malamute, malemute, Alaskan malamute
    250: "🐶",  # Siberian husky
    251: "🐶",  # dalmatian, coach dog, carriage dog
    252: "🐶",  # affenpinscher, monkey pinscher, monkey dog
    253: "🐶",  # basenji
    254: "🐶",  # pug, pug-dog
    255: "🐶",  # Leonberg
    256: "🐶",  # Newfoundland, Newfoundland dog
    257: "🐶",  # Great Pyrenees
    258: "🐶",  # Samoyed, Samoyede
    259: "🐶",  # Pomeranian
    260: "🐶",  # chow, chow chow
    261: "🐶",  # keeshond
    262: "🐶",  # Brabancon griffon
    263: "🐶",  # Pembroke, Pembroke Welsh corgi
    264: "🐶",  # Cardigan, Cardigan Welsh corgi
    265: "🐩",  # toy poodle
    266: "🐩",  # miniature poodle
    267: "🐩",  # standard poodle
    268: "🐶",  # Mexican hairless
    269: "🐺",  # timber wolf, grey wolf, gray wolf, Canis lupus
    270: "🐺",  # white wolf, Arctic wolf, Canis lupus tundrarum
    271: "🐺",  # red wolf, maned wolf, Canis rufus, Canis niger
    272: "🐺",  # coyote, prairie wolf, brush wolf, Canis latrans
    273: "🐺",  # dingo, warrigal, warragal, Canis dingo
    274: "🐺",  # dhole, Cuon alpinus
    275: "🐺",  # African hunting dog, hyena dog, Cape hunting dog, Lycaon pictus
    276: "🐺",  # hyena, hyaena
    277: "🦊",  # red fox, Vulpes vulpes
    278: "🦊",  # kit fox, Vulpes macrotis
    279: "🦊",  # Arctic fox, white fox, Alopex lagopus
    280: "🦊",  # grey fox, gray fox, Urocyon cinereoargenteus
    281: "🐱",  # tabby, tabby cat
    282: "🐱",  # tiger cat
    283: "🐱",  # Persian cat
    284: "🐱",  # Siamese cat, Siamese
    285: "🐱",  # Egyptian cat
    286: "🐆",  # cougar, puma, catamount, mountain lion, painter, panther, Felis concolor
    287: "🐆",  # lynx, catamount
    288: "🐆",  # leopard, Panthera pardus
    289: "🐆",  # snow leopard, ounce, Panthera uncia
    290: "🐆",  # jaguar, panther, Panthera onca, Felis onca
    291: "🦁",  # lion, king of beasts, Panthera leo
    292: "🐅",  # tiger, Panthera tigris
    293: "🐆",  # cheetah, chetah, Acinonyx jubatus
    294: "🐻",  # brown bear, bruin, Ursus arctos
    295: "🐻",  # American black bear, black bear, Ursus americanus, Euarctos americanus
    296: "🐻‍❄️",  # ice bear, polar bear, Ursus Maritimus, Thalarctos maritimus
    297: "🐻",  # sloth bear, Melursus ursinus, Ursus ursinus
    298: "🦦",  # mongoose
    299: "🦦",  # meerkat, mierkat
    300: "🐞",  # tiger beetle
    301: "🐞",  # ladybug, ladybeetle, lady beetle, ladybird, ladybird beetle
    302: "🐞",  # ground beetle, carabid beetle
    303: "🐞",  # long-horned beetle, longicorn, longicorn beetle
    304: "🐞",  # leaf beetle, chrysomelid
    305: "🐞",  # dung beetle
    306: "🐞",  # rhinoceros beetle
    307: "🐞",  # weevil
    308: "🪰",  # fly
    309: "🐝",  # bee
    310: "🐜",  # ant, emmet, pismire
    311: "🦗",  # grasshopper, hopper
    312: "🦗",  # cricket
    313: "🦗",  # walking stick, walkingstick, stick insect
    314: "🪳",  # cockroach, roach
    315: "🦗",  # mantis, mantid
    316: "🪰",  # cicada, cicala
    317: "🪰",  # leafhopper
    318: "🦋",  # lacewing, lacewing fly
    319: "🦋",  # dragonfly, darning needle, devil's darning needle, sewing needle, snake feeder, snake doctor, mosquito hawk, skeeter hawk
    320: "🦋",  # damselfly
    321: "🦋",  # admiral
    322: "🦋",  # ringlet, ringlet butterfly
    323: "🦋",  # monarch, monarch butterfly, milkweed butterfly, Danaus plexippus
    324: "🦋",  # cabbage butterfly
    325: "🦋",  # sulphur butterfly, sulfur butterfly
    326: "🦋",  # lycaenid, lycaenid butterfly
    327: "🌟",  # starfish, sea star
    328: "🦔",  # sea urchin
    329: "🦔",  # sea cucumber, holothurian
    330: "🐰",  # wood rabbit, cottontail, cottontail rabbit
    331: "🐰",  # hare
    332: "🐰",  # Angora, Angora rabbit
    333: "🐹",  # hamster
    334: "🦔",  # porcupine, hedgehog
    335: "🐿️",  # fox squirrel, eastern fox squirrel, Sciurus niger
    336: "🐹",  # marmot
    337: "🦫",  # beaver
    338: "🐹",  # guinea pig, Cavia cobaya
    339: "🐴",  # sorrel
    340: "🦓",  # zebra
    341: "🐷",  # hog, pig, grunter, squealer, Sus scrofa
    342: "🐗",  # wild boar, boar, Sus scrofa
    343: "🐗",  # warthog
    344: "🦛",  # hippopotamus, hippo, river horse, Hippopotamus amphibius
    345: "🐂",  # ox
    346: "🐃",  # water buffalo, water ox, Asiatic buffalo, Bubalus bubalis
    347: "🦬",  # bison
    348: "🐏",  # ram, tup
    349: "🐏",  # bighorn, bighorn sheep, cimarron, Rocky Mountain bighorn, Rocky Mountain sheep, Ovis canadensis
    350: "🐐",  # ibex, Capra ibex
    351: "🦌",  # hartebeest
    352: "🦌",  # impala, Aepyceros melampus
    353: "🦌",  # gazelle
    354: "🐪",  # Arabian camel, dromedary, Camelus dromedarius
    355: "🦙",  # llama
    356: "🦡",  # weasel
    357: "🦡",  # mink
    358: "🦡",  # polecat, fitch, foulmart, foumart, Mustela putorius
    359: "🦡",  # black-footed ferret, ferret, Mustela nigripes
    360: "🦦",  # otter
    361: "🦨",  # skunk, polecat, wood pussy
    362: "🦡",  # badger
    363: "🦦",  # armadillo
    364: "🦥",  # three-toed sloth, ai, Bradypus tridactylus
    365: "🦧",  # orangutan, orang, orangutang, Pongo pygmaeus
    366: "🦍",  # gorilla, Gorilla gorilla
    367: "🦧",  # chimpanzee, chimp, Pan troglodytes
    368: "🦧",  # gibbon, Hylobates lar
    369: "🦧",  # siamang, Hylobates syndactylus, Symphalangus syndactylus
    370: "🐒",  # guenon, guenon monkey
    371: "🐒",  # patas, hussar monkey, Erythrocebus patas
    372: "🦍",  # baboon
    373: "🐒",  # macaque
    374: "🐒",  # langur
    375: "🐒",  # colobus, colobus monkey
    376: "🐒",  # proboscis monkey, Nasalis larvatus
    377: "🐒",  # marmoset
    378: "🐒",  # capuchin, ringtail, Cebus capucinus
    379: "🐒",  # howler monkey, howler
    380: "🐒",  # titi, titi monkey
    381: "🐒",  # spider monkey, Ateles geoffroyi
    382: "🐒",  # squirrel monkey, Saimiri sciureus
    383: "🐒",  # Madagascar cat, ring-tailed lemur, Lemur catta
    384: "🐒",  # indri, indris, Indri indri, Indri brevicaudatus
    385: "🐘",  # Indian elephant, Elephas maximus
    386: "🐘",  # African elephant, Loxodonta africana
    387: "🦝",  # lesser panda, red panda, panda, bear cat, cat bear, Ailurus fulgens
    388: "🐼",  # giant panda, panda, panda bear, coon bear, Ailuropoda melanoleuca
    389: "🐟",  # barracouta, snoek
    390: "🐍",  # eel
    391: "🐟",  # coho, cohoe, coho salmon, blue jack, silver salmon, Oncorhynchus kisutch
    392: "🐠",  # rock beauty, Holocanthus tricolor
    393: "🐠",  # anemone fish
    394: "🐟",  # sturgeon
    395: "🐠",  # gar, garfish, garpike, billfish, Lepisosteus osseus
    396: "🐠",  # lionfish
    397: "🐡",  # puffer, pufferfish, blowfish, globefish
    398: "🧮",  # abacus
    399: "🧕",  # abaya
    400: "👨‍🎓",  # academic gown, academic robe, judge's robe
    401: "🪗",  # accordion, piano accordion, squeeze box
    402: "🎸",  # acoustic guitar
    403: "🚢",  # aircraft carrier, carrier, flattop, attack aircraft carrier
    404: "✈️",  # airliner
    405: "🎈",  # airship, dirigible
    406: "⛪",  # altar
    407: "🚑",  # ambulance
    408: "🚤",  # amphibian, amphibious vehicle
    409: "⏰",  # analog clock
    410: "🐝",  # apiary, bee house
    411: "🥼",  # apron
    412: "🗑️",  # ashcan, trash can, garbage can, wastebin, ash bin, ash-bin, ashbin, dustbin, trash barrel, trash bin
    413: "🔫",  # assault rifle, assault gun
    414: "🎒",  # backpack, back pack, knapsack, packsack, rucksack, haversack
    415: "🏢",  # bakery, bakeshop, bakehouse
    416: "🤸‍♀️",  # balance beam, beam
    417: "🎈",  # balloon
    418: "🖊️",  # ballpoint, ballpoint pen, ballpen, Biro
    419: "🩹",  # Band Aid
    420: "🪕",  # banjo
    421: "🪜",  # bannister, banister, balustrade, balusters, handrail
    422: "🏋️‍♂️",  # barbell
    423: "💺",  # barber chair
    424: "💈",  # barbershop
    425: "🏚️",  # barn
    426: "🌡️",  # barometer
    427: "🛢️",  # barrel, cask
    428: "🛒",  # barrow, garden cart, lawn cart, wheelbarrow
    429: "⚾",  # baseball
    430: "🏀",  # basketball
    431: "👶",  # bassinet
    432: "🎵",  # bassoon
    433: "🏊‍♀️",  # bathing cap, swimming cap
    434: "🛁",  # bath towel
    435: "🛁",  # bathtub, bathing tub, bath, tub
    436: "🚙",  # beach wagon, station wagon, wagon, estate car, beach waggon, station waggon, waggon
    437: "🏖️",  # beacon, lighthouse, beacon light, pharos
    438: "🧪",  # beaker
    439: "🎩",  # bearskin, busby, shako
    440: "🍺",  # beer bottle
    441: "🍺",  # beer glass
    442: "🔔",  # bell cote, bell cot
    443: "🍼",  # bib
    444: "🚲",  # bicycle-built-for-two, tandem bicycle, tandem
    445: "👙",  # bikini, two-piece
    446: "📁",  # binder, ring-binder
    447: "🔭",  # binoculars, field glasses, opera glasses
    448: "🐦",  # birdhouse
    449: "🏠",  # boathouse
    450: "🛷",  # bobsled, bobsleigh, bob
    451: "👔",  # bolo tie, bolo, bola tie, bola
    452: "👒",  # bonnet, poke bonnet
    453: "📚",  # bookcase
    454: "📚",  # bookshop, bookstore, bookstall
    455: "🔝",  # bottlecap
    456: "🏹",  # bow
    457: "🎀",  # bow tie, bow-tie, bowtie
    458: "🏵️",  # brass, memorial tablet, plaque
    459: "👙",  # brassiere, bra, bandeau
    460: "🚧",  # breakwater, groin, groyne, mole, bulwark, seawall, jetty
    461: "🛡️",  # breastplate, aegis, egis
    462: "🧹",  # broom
    463: "🪣",  # bucket, pail
    464: "🔒",  # buckle
    465: "🦺",  # bulletproof vest
    466: "🚆",  # bullet train, bullet
    467: "🥩",  # butcher shop, meat market
    468: "🚕",  # cab, hack, taxi, taxicab
    469: "🍲",  # caldron, cauldron
    470: "🕯️",  # candle, taper, wax light
    471: "💣",  # cannon
    472: "🛶",  # canoe
    473: "🔪",  # can opener, tin opener
    474: "🧥",  # cardigan
    475: "🪞",  # car mirror
    476: "🎠",  # carousel, carrousel, merry-go-round, roundabout, whirligig
    477: "🧰",  # carpenter's kit, tool kit
    478: "📦",  # carton
    479: "🚗",  # car wheel
    480: "🏧",  # cash machine, cash dispenser, automated teller machine, automatic teller machine, automated teller, automatic teller, ATM
    481: "📼",  # cassette
    482: "📼",  # cassette player
    483: "🏰",  # castle
    484: "⛵",  # catamaran
    485: "💿",  # CD player
    486: "🎻",  # cello, violoncello
    487: "📱",  # cellular telephone, cellular phone, cellphone, cell, mobile phone
    488: "⛓️",  # chain
    489: "🔗",  # chainlink fence
    490: "🛡️",  # chain mail, ring mail, mail, chain armor, chain armour, ring armor, ring armour
    491: "⛓️",  # chain saw, chainsaw
    492: "🧰",  # chest
    493: "🚪",  # chiffonier, commode
    494: "🔔",  # chime, bell, gong
    495: "🏺",  # china cabinet, china closet
    496: "🧦",  # Christmas stocking
    497: "⛪",  # church, church building
    498: "🎦",  # cinema, movie theater, movie theatre, movie house, picture palace
    499: "🔪",  # cleaver, meat cleaver, chopper
    500: "🏚️",  # cliff dwelling
    501: "🧥",  # cloak
    502: "👞",  # clog, geta, patten, sabot
    503: "🍸",  # cocktail shaker
    504: "☕",  # coffee mug
    505: "☕",  # coffeepot
    506: "🌀",  # coil, spiral, volute, whorl, helix
    507: "🔒",  # combination lock
    508: "⌨️",  # computer keyboard, keypad
    509: "🍬",  # confectionery, confectionary, candy store
    510: "🚢",  # container ship, containership, container vessel
    511: "🚗",  # convertible
    512: "🍾",  # corkscrew, bottle screw
    513: "🎺",  # cornet, horn, trumpet, trump
    514: "👢",  # cowboy boot
    515: "🤠",  # cowboy hat, ten-gallon hat
    516: "🛏️",  # cradle
    517: "🏗️",  # crane
    518: "🏍️",  # crash helmet
    519: "📦",  # crate
    520: "🛏️",  # crib, cot
    521: "🍲",  # Crock Pot
    522: "🏑",  # croquet ball
    523: "🦯",  # crutch
    524: "🛡️",  # cuirass
    525: "🏞️",  # dam, dike, dyke
    526: "🖥️",  # desk
    527: "🖥️",  # desktop computer
    528: "📞",  # dial telephone, dial phone
    529: "👶",  # diaper, nappy, napkin
    530: "⏲️",  # digital clock
    531: "⌚",  # digital watch
    532: "🍽️",  # dining table, board
    533: "🧼",  # dishrag, dishcloth
    534: "🧼",  # dishwasher, dish washer, dishwashing machine
    535: "🚗",  # disk brake, disc brake
    536: "🚢",  # dock, dockage, docking facility
    537: "🛷",  # dogsled, dog sled, dog sleigh
    538: "🏰",  # dome
    539: "🚪",  # doormat, welcome mat
    540: "🛢️",  # drilling platform, offshore rig
    541: "🥁",  # drum, membranophone, tympan
    542: "🍗",  # drumstick
    543: "🏋️",  # dumbbell
    544: "🥘",  # Dutch oven
    545: "🌀",  # electric fan, blower
    546: "🎸",  # electric guitar
    547: "🚂",  # electric locomotive
    548: "📺",  # entertainment center
    549: "✉️",  # envelope
    550: "☕",  # espresso maker
    551: "💄",  # face powder
    552: "🧣",  # feather boa, boa
    553: "🗄️",  # file, file cabinet, filing cabinet
    554: "🚒",  # fireboat
    555: "🚒",  # fire engine, fire truck
    556: "🔥",  # fire screen, fireguard
    557: "🚩",  # flagpole, flagstaff
    558: "🎶",  # flute, transverse flute
    559: "🪑",  # folding chair
    560: "🏈",  # football helmet
    561: "🚜",  # forklift
    562: "⛲",  # fountain
    563: "✒️",  # fountain pen
    564: "🛏️",  # four-poster
    565: "🚃",  # freight car
    566: "📯",  # French horn, horn
    567: "🍳",  # frying pan, frypan, skillet
    568: "🧥",  # fur coat
    569: "🚚",  # garbage truck, dustcart
    570: "🥽",  # gasmask, respirator, gas helmet
    571: "⛽",  # gas pump, gasoline pump, petrol pump, island dispenser
    572: "🍷",  # goblet
    573: "🏎️",  # go-kart
    574: "⛳",  # golf ball
    575: "🏌️",  # golfcart, golf cart
    576: "🚣",  # gondola
    577: "🔔",  # gong, tam-tam
    578: "👗",  # gown
    579: "🎹",  # grand piano, grand
    580: "🏡",  # greenhouse, nursery, glasshouse
    581: "🚗",  # grille, radiator grille
    582: "🏪",  # grocery store, grocery, food market, market
    583: "🔪",  # guillotine
    584: "📎",  # hair slide
    585: "💈",  # hair spray
    586: "🚜",  # half track
    587: "🔨",  # hammer
    588: "🧺",  # hamper
    589: "💇",  # hand blower, blow dryer, blow drier, hair dryer, hair drier
    590: "💻",  # hand-held computer, hand-held microcomputer
    591: "🧻",  # handkerchief, hankie, hanky, hankey
    592: "💾",  # hard disc, hard disk, fixed disk
    593: "🎶",  # harmonica, mouth organ, harp, mouth harp
    594: "🎵",  # harp
    595: "🚜",  # harvester, reaper
    596: "🪓",  # hatchet
    597: "🔫",  # holster
    598: "📺",  # home theater, home theatre
    599: "🐝",  # honeycomb
    600: "🪝",  # hook, claw
    601: "👗",  # hoopskirt, crinoline
    602: "🏋️",  # horizontal bar, high bar
    603: "🐎",  # horse cart, horse-cart
    604: "⏳",  # hourglass
    605: "📱",  # iPod
    606: "🔥",  # iron, smoothing iron
    607: "🎃",  # jack-o'-lantern
    608: "👖",  # jean, blue jean, denim
    609: "🚙",  # jeep, landrover
    610: "👕",  # jersey, T-shirt, tee shirt
    611: "🧩",  # jigsaw puzzle
    612: "🛺",  # jinrikisha, ricksha, rickshaw
    613: "🕹️",  # joystick
    614: "👘",  # kimono
    615: "🛡️",  # knee pad
    616: "🪢",  # knot
    617: "🥼",  # lab coat, laboratory coat
    618: "🥄",  # ladle
    619: "💡",  # lampshade, lamp shade
    620: "💻",  # laptop, laptop computer
    621: "🌾",  # lawn mower, mower
    622: "📷",  # lens cap, lens cover
    623: "🔪",  # letter opener, paper knife, paperknife
    624: "📚",  # library
    625: "🛶",  # lifeboat
    626: "🔥",  # lighter, light, igniter, ignitor
    627: "🚘",  # limousine, limo
    628: "🚢",  # liner, ocean liner
    629: "💋",  # lipstick, lip rouge
    630: "🥿",  # Loafer
    631: "🧴",  # lotion
    632: "🔊",  # loudspeaker, speaker, speaker unit, loudspeaker system, speaker system
    633: "🔍",  # loupe, jeweler's loupe
    634: "🏭",  # lumbermill, sawmill
    635: "🧭",  # magnetic compass
    636: "📮",  # mailbag, postbag
    637: "📫",  # mailbox, letter box
    638: "🩱",  # maillot
    639: "🩱",  # maillot, tank suit
    640: "🕳️",  # manhole cover
    641: "🎊",  # maraca
    642: "🎵",  # marimba, xylophone
    643: "😷",  # mask
    644: "🪚",  # matchstick
    645: "🎪",  # maypole
    646: "🌀",  # maze, labyrinth
    647: "🥣",  # measuring cup
    648: "💊",  # medicine chest, medicine cabinet
    649: "🪨",  # megalith, megalithic structure
    650: "🎤",  # microphone, mike
    651: "🍳",  # microwave, microwave oven
    652: "🪖",  # military uniform
    653: "🥛",  # milk can
    654: "🚌",  # minibus
    655: "👗",  # miniskirt, mini
    656: "🚐",  # minivan
    657: "🚀",  # missile
    658: "🧤",  # mitten
    659: "🥣",  # mixing bowl
    660: "🏠",  # mobile home, manufactured home
    661: "🚗",  # Model T
    662: "💻",  # modem
    663: "🕌",  # monastery
    664: "🖥️",  # monitor
    665: "🛵",  # moped
    666: "🍲",  # mortar
    667: "🎓",  # mortarboard
    668: "🕌",  # mosque
    669: "🛌",  # mosquito net
    670: "🛵",  # motor scooter, scooter
    671: "🚵",  # mountain bike, all-terrain bike, off-roader
    672: "⛺",  # mountain tent
    673: "🖱️",  # mouse, computer mouse
    674: "🪤",  # mousetrap
    675: "🚚",  # moving van
    676: "🐾",  # muzzle
    677: "🔩",  # nail
    678: "🤕",  # neck brace
    679: "💎",  # necklace
    680: "🍼",  # nipple
    681: "💻",  # notebook, notebook computer
    682: "🗿",  # obelisk
    683: "🎷",  # oboe, hautboy, hautbois
    684: "🎵",  # ocarina, sweet potato
    685: "🚗",  # odometer, hodometer, mileometer, milometer
    686: "🛢️",  # oil filter
    687: "🎹",  # organ, pipe organ
    688: "🔬",  # oscilloscope, scope, cathode-ray oscilloscope, CRO
    689: "👗",  # overskirt
    690: "🛒",  # oxcart
    691: "🎭",  # oxygen mask
    692: "📦",  # packet
    693: "🛶",  # paddle, boat paddle
    694: "🎡",  # paddlewheel, paddle wheel
    695: "🔒",  # padlock
    696: "🖌️",  # paintbrush
    697: "👚",  # pajama, pyjama, pj's, jammies
    698: "🏰",  # palace
    699: "🎼",  # panpipe, pandean pipe, syrinx
    700: "🧻",  # paper towel
    701: "🪂",  # parachute, chute
    702: "🏋️",  # parallel bars, bars
    703: "🪑",  # park bench
    704: "🅿️",  # parking meter
    705: "🚗",  # passenger car, coach, carriage
    706: "🏡",  # patio, terrace
    707: "📞",  # pay-phone, pay-station
    708: "🧱",  # pedestal, plinth, footstall
    709: "📦",  # pencil box, pencil case
    710: "✏️",  # pencil sharpener
    711: "🌺",  # perfume, essence
    712: "🧫",  # Petri dish
    713: "🖨️",  # photocopier
    714: "🎸",  # pick, plectrum, plectron
    715: "🪖",  # pickelhaube
    716: "🏡",  # picket fence, paling
    717: "🚚",  # pickup, pickup truck
    718: "🏖️",  # pier
    719: "🐖",  # piggy bank, penny bank
    720: "💊",  # pill bottle
    721: "🛌",  # pillow
    722: "🏓",  # ping-pong ball
    723: "🌀",  # pinwheel
    724: "🏴‍☠️",  # pirate, pirate ship
    725: "🍶",  # pitcher, ewer
    726: "✈️",  # plane, carpenter's plane, woodworking plane
    727: "🪐",  # planetarium
    728: "🛍️",  # plastic bag
    729: "🍽️",  # plate rack
    730: "🚜",  # plow, plough
     731: "🚽",  # plunger, plumber's helper
     732: "📸",  # Polaroid camera, Polaroid Land camera
     733: "🏑",  # pole
     734: "🚓",  # police van, police wagon, paddy wagon, patrol wagon, wagon, black Maria
     735: "🧥",  # poncho
     736: "🎱",  # pool table, billiard table, snooker table
     737: "🥤",  # pop bottle, soda bottle
     738: "🪴",  # pot, flowerpot
     739: "🏺",  # potter's wheel
     740: "🛠️",  # power drill
     741: "🧎",  # prayer rug, prayer mat
     742: "🖨️",  # printer
     743: "🏛️",  # prison, prison house
     744: "🚀",  # projectile, missile
     745: "📽️",  # projector
     746: "🏒",  # puck, hockey puck
     747: "🥊",  # punching bag, punch bag, punching ball, punchball
     748: "👜",  # purse
     749: "✒️",  # quill, quill pen
     750: "🛌",  # quilt, comforter, comfort, puff
     751: "🏎️",  # racer, race car, racing car
     752: "🎾",  # racket, racquet
     753: "🔥",  # radiator
     754: "📻",  # radio, wireless
     755: "🔭",  # radio telescope, radio reflector
     756: "🛢️",  # rain barrel
     757: "🚐",  # recreational vehicle, RV, R.V.
     758: "🎣",  # reel
     759: "📷",  # reflex camera
     760: "🧊",  # refrigerator, icebox
     761: "🔦",  # remote control, remote
     762: "🍽️",  # restaurant, eating house, eating place, eatery
     763: "🔫",  # revolver, six-gun, six-shooter
     764: "🔫",  # rifle
     765: "🪑",  # rocking chair, rocker
     766: "🍗",  # rotisserie
     767: "✏️",  # rubber eraser, rubber, pencil eraser
     768: "🏉",  # rugby ball
     769: "📏",  # rule, ruler
     770: "👟",  # running shoe
     771: "🔐",  # safe
     772: "🧷",  # safety pin
     773: "🧂",  # saltshaker, salt shaker
     774: "🩴",  # sandal
     775: "🧣",  # sarong
     776: "🎷",  # sax, saxophone
     777: "🗡️",  # scabbard
     778: "⚖️",  # scale, weighing machine
     779: "🚌",  # school bus
     780: "⛵",  # schooner
     781: "📊",  # scoreboard
     782: "🖥️",  # screen, CRT screen
     783: "🔩",  # screw
     784: "🪛",  # screwdriver
     785: "🚗",  # seat belt, seatbelt
     786: "🧵",  # sewing machine
     787: "🛡️",  # shield, buckler
     788: "👠",  # shoe shop, shoe-shop, shoe store
     789: "🚪",  # shoji
     790: "🛒",  # shopping basket
     791: "🛒",  # shopping cart
     792: "🪓",  # shovel
     793: "🚿",  # shower cap
     794: "🚿",  # shower curtain
     795: "⛷️",  # ski
     796: "🎿",  # ski mask
     797: "🎒",  # sleeping bag
     798: "📏",  # slide rule, slipstick
     799: "🚪",  # sliding door
     800: "🎰",  # slot, one-armed bandit
     801: "🤿",  # snorkel
     802: "🌨️",  # snowmobile
     803: "🚜",  # snowplow, snowplough
     804: "🧼",  # soap dispenser
     805: "⚽",  # soccer ball
     806: "🧦",  # sock
     807: "☀️",  # solar dish, solar collector, solar furnace
     808: "🎩",  # sombrero
     809: "🍲",  # soup bowl
     810: "⌨️",  # space bar
     811: "🔥",  # space heater
     812: "🚀",  # space shuttle
     813: "🥄",  # spatula
     814: "🚤",  # speedboat
     815: "🕸️",  # spider web, spider's web
     816: "🧶",  # spindle
     817: "🏎️",  # sports car, sport car
     818: "💡",  # spotlight, spot
     819: "🎭",  # stage
     820: "🚂",  # steam locomotive
     821: "🌉",  # steel arch bridge
     822: "🥁",  # steel drum
     823: "🩺",  # stethoscope
     824: "🧣",  # stole
     825: "🧱",  # stone wall
     826: "⏱️",  # stopwatch, stop watch
     827: "🍳",  # stove
     828: "🍳",  # strainer
     829: "🚋",  # streetcar, tram, tramcar, trolley, trolley car
     830: "🚑",  # stretcher
     831: "🛋️",  # studio couch, day bed
     832: "🕌",  # stupa, tope
     833: "🚢",  # submarine, pigboat, sub, U-boat
     834: "👔",  # suit, suit of clothes
     835: "⏳",  # sundial
     836: "🕶️",  # sunglass
     837: "🕶️",  # sunglasses, dark glasses, shades
     838: "🧴",  # sunscreen, sunblock, sun blocker
     839: "🌉",  # suspension bridge
     840: "🧹",  # swab, swob, mop
     841: "👕",  # sweatshirt
     842: "🩳",  # swimming trunks, bathing trunks
     843: "🎠",  # swing
     844: "🔌",  # switch, electric switch, electrical switch
     845: "💉",  # syringe
     846: "💡",  # table lamp
     847: "🚜",  # tank, army tank, armored combat vehicle, armoured combat vehicle
     848: "📼",  # tape player
     849: "☕",  # teapot
     850: "🧸",  # teddy, teddy bear
     851: "📺",  # television, television system
     852: "🎾",  # tennis ball
     853: "🏠",  # thatch, thatched roof
     854: "🎭",  # theater curtain, theatre curtain
     855: "🧵",  # thimble
     856: "🚜",  # thresher, thrasher, threshing machine
     857: "👑",  # throne
     858: "🏠",  # tile roof
     859: "🍞",  # toaster
     860: "🏪",  # tobacco shop, tobacconist shop, tobacconist
     861: "🚽",  # toilet seat
     862: "🔦",  # torch
     863: "🪱",  # totem pole
     864: "🚚",  # tow truck, tow car, wrecker
     865: "🏬",  # toyshop
     866: "🚜",  # tractor
     867: "🚛",  # trailer truck, tractor trailer, trucking rig, rig, articulated lorry, semi
     868: "🍽️",  # tray
     869: "🧥",  # trench coat
     870: "🚲",  # tricycle, trike, velocipede
     871: "⛵",  # trimaran
     872: "📸",  # tripod
     873: "🏛️",  # triumphal arch
     874: "🚎",  # trolleybus, trolley coach, trackless trolley
     875: "🎺",  # trombone
     876: "🛁",  # tub, vat
     877: "🚧",  # turnstile
     878: "⌨️",  # typewriter keyboard
     879: "☂️",  # umbrella
     880: "🚲",  # unicycle, monocycle
     881: "🎹",  # upright, upright piano
     882: "🧹",  # vacuum, vacuum cleaner
     883: "🏺",  # vase
     884: "🚪",  # vault
     885: "🧣",  # velvet
     886: "🥤",  # vending machine
     887: "👗",  # vestment
     888: "🌉",  # viaduct
     889: "🎻",  # violin, fiddle
     890: "🏐",  # volleyball
     891: "🧇",  # waffle iron
     892: "⏰",  # wall clock
     893: "👛",  # wallet, billfold, notecase, pocketbook
     894: "🚪",  # wardrobe, closet, press
     895: "✈️",  # warplane, military plane
     896: "🚰",  # washbasin, handbasin, washbowl, lavabo, wash-hand basin
     897: "🧼",  # washer, automatic washer, washing machine
     898: "🍶",  # water bottle
     899: "🍶",  # water jug
     900: "🌊",  # water tower
     901: "🍶",  # whiskey jug
     902: "📣",  # whistle
     903: "💇",  # wig
     904: "🪟",  # window screen
     905: "🪟",  # window shade
     906: "👔",  # Windsor tie
     907: "🍷",  # wine bottle
     908: "✈️",  # wing
     909: "🍳",  # wok
     910: "🥄",  # wooden spoon
     911: "🧶",  # wool, woolen, woollen
     912: "🚧",  # worm fence, snake fence, snake-rail fence, Virginia fence
     913: "🚢",  # wreck
     914: "⛵",  # yawl
     915: "🏕️",  # yurt
     916: "💻",  # web site, website, internet site, site
     917: "📚",  # comic book
     918: "✏️",  # crossword puzzle, crossword
     919: "🚏",  # street sign
     920: "🚦",  # traffic light, traffic signal, stoplight
     921: "📚",  # book jacket, dust cover, dust jacket, dust wrapper
     922: "📖",  # menu
     923: "🍽️",  # plate
     924: "🥑",  # guacamole
     925: "🍲",  # consomme
     926: "🍲",  # hot pot, hotpot
     927: "🍰",  # trifle
     928: "🍨",  # ice cream, icecream
     929: "🍦",  # ice lolly, lolly, lollipop, popsicle
     930: "🥖",  # French loaf
     931: "🥯",  # bagel, beigel
     932: "🥨",  # pretzel
     933: "🍔",  # cheeseburger
     934: "🌭",  # hotdog, hot dog, red hot
     935: "🥔",  # mashed potato
     936: "🥬",  # head cabbage
     937: "🥦",  # broccoli
     938: "🥦",  # cauliflower
     939: "🥒",  # zucchini, courgette
     940: "🥒",  # spaghetti squash
     941: "🥔",  # acorn squash
     942: "🥔",  # butternut squash
     943: "🥒",  # cucumber, cuke
     944: "🥦",  # artichoke, globe artichoke
     945: "🌶️",  # bell pepper
     946: "🥦",  # cardoon
     947: "🍄",  # mushroom
     948: "🍏",  # Granny Smith
     949: "🍓",  # strawberry
     950: "🍊",  # orange
     951: "🍋",  # lemon
     952: "🍇",  # fig
     953: "🍍",  # pineapple, ananas
     954: "🍌",  # banana
     955: "🍈",  # jackfruit, jak, jack
     956: "🍎",  # custard apple
     957: "🍇",  # pomegranate
     958: "🌾",  # hay
     959: "🍝",  # carbonara
     960: "🍫",  # chocolate sauce, chocolate syrup
     961: "🍞",  # dough
     962: "🍞",  # meat loaf, meatloaf
     963: "🍕",  # pizza, pizza pie
     964: "🥧",  # potpie
     965: "🌯",  # burrito
     966: "🍷",  # red wine
     967: "☕",  # espresso
     968: "☕",  # cup
     969: "🥛",  # eggnog
     970: "⛰️",  # alp
     971: "💭",  # bubble
     972: "🏞️",  # cliff, drop, drop-off
     973: "🏝️",  # coral reef
     974: "⛲",  # geyser
     975: "🏖️",  # lakeside, lakeshore
     976: "🏞️",  # promontory, headland, head, foreland
     977: "🏖️",  # sandbar, sand bar
     978: "🏖️",  # seashore, coast, seacoast, sea-coast
     979: "🏞️",  # valley, vale
     980: "🌋",  # volcano
     981: "⚾",  # ballplayer, baseball player
     982: "🤵",  # groom, bridegroom
     983: "🤿",  # scuba diver
     984: "🌾",  # rapeseed
     985: "🌼",  # daisy
     986: "🌼",  # yellow lady's slipper, yellow lady-slipper, Cypripedium calceolus, Cypripedium parviflorum
     987: "🌽",  # corn
     988: "🌰",  # acorn
     989: "🌹",  # hip, rose hip, rosehip
     990: "🌰",  # buckeye, horse chestnut, conker
     991: "🍄",  # coral fungus
     992: "🍄",  # agaric
     993: "🍄",  # gyromitra
     994: "🍄",  # stinkhorn, carrion fungus
     995: "🍄",  # earthstar
     996: "🍄",  # hen-of-the-woods, hen of the woods, Polyporus frondosus, Grifola frondosa
     997: "🍄",  # bolete
     998: "🌾",  # ear, spike, capitulum
     999: "🧻"  # toilet tissue, toilet paper, bathroom tissue
    }



import torch.nn as nn
import torch

import logging
from vit_prisma.prisma_tools import HookPoint
from vit_prisma.configs.HookedViTConfig import HookedViTConfig

from typing import Union, Dict, Optional, List, Tuple

from jaxtyping import Float, Int

import numpy as np

import einops

from vit_prisma.prisma_tools import FactoredMatrix
from fancy_einsum import einsum

import torch.nn.functional as F



class Attention(nn.Module):

    def __init__(
            self,
            cfg: Union[Dict, HookedViTConfig],
            layer_id: Optional[int] = None,
    ):
        super().__init__()
        if isinstance(cfg, Dict):
            cfg = HookedViTConfig.from_dict(cfg)

        self.cfg = cfg

        # Initialize parameters
        self.W_Q = nn.Parameter(
            torch.empty(
                self.cfg.n_heads,
                self.cfg.d_model,
                self.cfg.d_head,
                dtype = self.cfg.dtype
            )
        )
        self.W_K = nn.Parameter(
            torch.empty(
                self.cfg.n_heads,
                self.cfg.d_model,
                self.cfg.d_head,
                dtype = self.cfg.dtype
            )
        )
        self.W_V = nn.Parameter(
            torch.empty(
                self.cfg.n_heads,
                self.cfg.d_model,
                self.cfg.d_head,
                dtype = self.cfg.dtype
            )
        )
        self.W_O = nn.Parameter(
            torch.empty(
                self.cfg.n_heads,
                self.cfg.d_head,
                self.cfg.d_model,
                dtype = self.cfg.dtype
            )
        )

        # Initialize biases
        self.b_Q = nn.Parameter(
            torch.zeros(self.cfg.n_heads, self.cfg.d_head, dtype=self.cfg.dtype)
        )
        self.b_K = nn.Parameter(
            torch.zeros(self.cfg.n_heads, self.cfg.d_head, dtype=self.cfg.dtype)
        )
        self.b_V = nn.Parameter(
            torch.zeros(self.cfg.n_heads, self.cfg.d_head, dtype=self.cfg.dtype)
        )
        self.b_O = nn.Parameter(torch.zeros(self.cfg.d_model, dtype=self.cfg.dtype))


        # Add hook points
        self.hook_k = HookPoint()  # [batch, pos, head_index, d_head]
        self.hook_q = HookPoint()  # [batch, pos, head_index, d_head]
        self.hook_v = HookPoint()  # [batch, pos, head_index, d_head]
        self.hook_z = HookPoint()  # [batch, pos, head_index, d_head]
        self.hook_attn_scores = HookPoint()  # [batch, head_index, query_pos, key_pos]
        self.hook_pattern = HookPoint()  # [batch, head_index, query_pos, key_pos]
        self.hook_result = HookPoint()  # [batch, pos, head_index, d_model]

        self.layer_id = layer_id

        # Note to Sonia: check this.
        # attn_scale is a constant that we divide the attention scores by pre-softmax. I'm not entirely sure why it matters, but it's probably a mix of softmax not being scale invariant and numerical stability?
        if self.cfg.use_attn_scale:
            self.attn_scale = np.sqrt(self.cfg.d_head)
        else:
            self.attn_scale = 1.0

    @property
    def OV(self) -> FactoredMatrix:
        """
        OV-Circuit, as defined in A Mathematical Framework. Because there's no non-linearity between the value vector and the output of the layer, the output is purely determined by the matrix W_OV = W_V @ W_O, and not W_V or W_O individually. (Mathematically, for a single head, output == pattern @ residual @ W_V @ W_O, see the glossary for more)

        Done in the order W_V, W_O because the paper uses left-multiplying weight matrices, and TransformerLens uses right-multiplying, sorry!

        Returns a FactoredMatrix, with left matrix W_V [head_index, d_model, d_head] and right matrix W_O [head_index, d_head, d_model] - this is a low rank factorisation of the underlying [head_index, d_model, d_model]. FactoredMatrix has helper functions to deal with these large matrices efficiently. To get the OV circuit of a head k, attn.OV[k] works.
        """
        return FactoredMatrix(self.W_V, self.W_O)

    @property
    def QK(self) -> FactoredMatrix:
        """
        QK-Circuit, as defined in A Mathematical Framework. Because there's no non-linearity in the key-query dot product, the output is purely determined by the matrix W_QK = W_Q.T @ W_K, and not W_Q or W_K individually. (Mathematically, for a single head, pattern = destination_residual.T @ W_Q.T @ W_K @ source-residual, see the glossary for more).

        Done in the order Q on the left, K on the right, because the pattern has dimensions [destination_pos, source_pos]

        Returns a FactoredMatrix, with left matrix W_Q [head_index, d_model, d_head] and right matrix W_K.T [head_index, d_head, d_model] - this is a low rank factorisation of the underlying [head_index, d_model, d_model] matrix. FactoredMatrix has helper functions to deal with these large matrices efficiently. To get the QK circuit of a head k, attn.QK[k] works.
        """
        W_K_transpose = einops.rearrange(
            self.W_K, "head_index d_model d_head -> head_index d_head d_model"
        )
        return FactoredMatrix(self.W_Q, W_K_transpose)
    
    def forward(
            self,
            query_input: Union[
                Float[torch.Tensor, "batch pos d_model"],
                Float[torch.Tensor, "batch pos head_index d_model"],
            ],
            key_input: Union[
                Float[torch.Tensor, "batch pos d_model"],
                Float[torch.Tensor, "batch pos head_index d_model"],
            ],
            value_input: Union[
                Float[torch.Tensor, "batch pos d_model"],
                Float[torch.Tensor, "batch pos head_index d_model"],
            ]
    ) -> Float[torch.Tensor, "batch pos d_model"]:
        
        q, k, v  = self.calculate_qkv_matrices(query_input, key_input, value_input)

        attn_scores = self.calculate_attn_scores(q, k)
        attn_scores = self.hook_attn_scores(attn_scores)

        pattern = F.softmax(attn_scores, dim=-1) # where do I do normalization? 
        pattern = torch.where(torch.isnan(pattern), torch.zeros_like(pattern), pattern)
        pattern = self.hook_pattern(pattern)

        pattern = pattern.to(self.cfg.dtype)
        z = self.calculate_z_scores(v, pattern)

        if not self.cfg.use_attn_result:
            out = (
                (
                    einsum(
                        "batch pos head_index d_head, \
                        head_index d_head d_model -> \
                        batch pos d_model",
                        z,
                        self.W_O,
                    )
                )
                + self.b_O
            )
        else: 
            # Explicitly calculate the attention result so it can be accessed by a hook.
            # Off by default to not eat through GPU memory.
            result = self.hook_result(
                einsum(
                    "batch pos head_index d_head, \
                    head_index d_head d_model -> \
                    batch pos head_index d_model",
                    z,
                    self.W_O,
                )
            )
            out = (
                einops.reduce(result, "batch pos head_index d_model -> batch pos d_model", "sum")
                + self.b_O
            )
        return out

    def calculate_qkv_matrices(
            self,
            query_input: Union[
                Float[torch.Tensor, "batch pos d_model"],
                Float[torch.Tensor, "batch pos head_index d_model"],
            ],
            key_input: Union[
                Float[torch.Tensor, "batch pos d_model"],
                Float[torch.Tensor, "batch pos head_index d_model"],
            ],
            value_input: Union[
                Float[torch.Tensor, "batch pos d_model"],
                Float[torch.Tensor, "batch pos head_index d_model"],
            ]
    ) -> Tuple[
        Float[torch.Tensor, "batch pos head_index d_head"],
        Float[torch.Tensor, "batch pos head_index d_head"],
        Float[torch.Tensor, "batch pos head_index d_head"],
    ]:
        """
        Calculate the Q, K, V matrices for the attention layer. This is done by multiplying the input by the weight matrices and adding the biases.

        Returns a tuple of (Q, K, V) matrices, each of shape [batch, pos, head_index, d_head]
        """

        if self.cfg.use_split_qkv_input or self.cfg.use_attn_in:
            qkv_einops_string = "batch pos head_index d_model"
        else:
            qkv_einops_string = "batch pos d_model"


        q = self.hook_q(
            einsum(
                f"{qkv_einops_string}, head_index d_model d_head \
                -> batch pos head_index d_head",
                query_input,
                self.W_Q,
            )
            + self.b_Q
        )  # [batch, pos, head_index, d_head]
        k = self.hook_k(
            einsum(
                f"{qkv_einops_string}, head_index d_model d_head \
                -> batch pos head_index d_head",
                key_input,
                self.W_K,
            )
            + self.b_K
        )  # [batch, pos, head_index, d_head]
        v = self.hook_v(
            einsum(
                f"{qkv_einops_string}, head_index d_model d_head \
                -> batch pos head_index d_head",
                value_input,
                self.W_V,
            )
            + self.b_V
        )  # [batch, pos, head_index, d_head]
        return q, k, v
    
    def calculate_attn_scores(
            self,
            q: Float[torch.Tensor, "batch pos head_index d_head"],
            k: Float[torch.Tensor, "batch pos head_index d_head"],
    ) -> Float[torch.Tensor, "batch head_index query_pos key_pos"]:
        """
        Calculate the attention scores for the attention layer. This is done by multiplying the Q and K matrices together, and dividing by the square root of the dimension of the key vectors.

        Returns a tensor of shape [batch, head_index, query_pos, key_pos]
        """
        attn_scores = einsum(
            "batch query_pos head_index d_head, batch key_pos head_index d_head -> batch head_index query_pos key_pos",
            q,
            k,
        )
        attn_scores = attn_scores / self.attn_scale
        return attn_scores
    
    def calculate_z_scores(
            self,
            v: Float[torch.Tensor, "batch key_pos head_index d_head"],
            pattern: Float[torch.Tensor, "batch head_index query_pos key_pos"],
    ) -> Float[torch.Tensor, "batch query_pos head_index d_head"]:
        z = self.hook_z(
            einsum(
                "batch key_pos head_index d_head, \
                batch head_index query_pos key_pos -> \
                batch query_pos head_index d_head",
                v,
                pattern,
            )
        )
        return z



import numpy as np
import torch
from torch.utils.data import Dataset, Subset
from sklearn.model_selection import train_test_split
import os


class DSpritesDataset(Dataset):
    def __init__(self, data_path= '/network/scratch/s/sonia.joseph/datasets/dsprites.npz'):
        # Load the dataset
        data = np.load(data_path, allow_pickle=True, encoding='latin1')

        self.images = data['imgs']
        self.labels =  data['latents_values'][:,1]

    def __len__(self):
        return len(self.images)

    def __getitem__(self, idx):
        # Convert image from numpy array to PyTorch tensor and add channel dimension
        image = torch.tensor(self.images[idx], dtype=torch.float32).unsqueeze(0)
        label = torch.tensor(self.labels[idx]-1, dtype=torch.int64)
        return image, label
    

def train_test_dataset(dataset, test_split=0.25):
    train_idx, val_idx = train_test_split(list(range(len(dataset))), test_size=test_split)
    datasets = {}
    datasets['train'] = Subset(dataset, train_idx)
    datasets['test'] = Subset(dataset, val_idx)
    return datasets


# Example to load the data with train and test split
## dsprites_dataset = DSpritesDataset(data_path)
## dsprites_datasets = train_test_dataset(dsprites_dataset, test_split=0.25)



"""
Prisma Repo
By Sonia Joseph

Copyright (c) Sonia Joseph. All rights reserved.

Inspired by TransformerLens. Some functions have been adapted from the TransformerLens project.
For more information on TransformerLens, visit: https://github.com/neelnanda-io/TransformerLens
"""

from typing import Dict, Optional, Tuple, Union, Callable, Sequence, List

from jaxtyping import Float, Int

import torch.nn as nn 
from vit_prisma.prisma_tools.hook_point import HookPoint

import logging

from contextlib import contextmanager


NamesFilter = Optional[Union[Callable[[str], bool], Sequence[str]]]

class HookedRootModule(nn.Module):

    def __init__(self, *args):
        super().__init__()

        self.is_caching = False
        self.context_level = 0
    
    def setup(self):
        """
        Sets up model.

        This function must be called in the model's `__init__` method AFTER defining all layers. It
        adds a parameter to each module containing its name, and builds a dictionary mapping module
        names to the module instances. It also initializes a hook dictionary for modules of type
        "HookPoint".
        """
        self.mod_dict = {}
        self.hook_dict: Dict[str, HookPoint] = {}
        for name, module in self.named_modules():
            if name == "":
                continue
            module.name = name
            self.mod_dict[name] = module
            if isinstance(module, HookPoint):
                self.hook_dict[name] = module
        
    def hook_points(self):
        return self.hook_dict.values()
    
    def remove_all_hook_fns(
        self, dir="both", including_permanent=False, level=None
    ):
        for hp in self.hook_points():
            hp.remove_hooks(dir, including_permanent, level)
    
    def clear_context(self):
        for hp in self.hook_points():
            hp.clear_context()
    
    def reset_hooks(
            self, clear_contexts=True, direction="both", including_permanent=False, level=None
    ) -> None:
        if clear_contexts:
            self.clear_context()
        self.remove_all_hook_fns(direction, including_permanent, level)
        self.is_caching = False
    
    def check_and_add_hook(self, hook_point, hook_point_name, hook, dir="fwd", is_permanent=False, level=None, prepend=False) -> None:
        self.check_hooks_to_add(
            hook_point,
            hook_point_name,
            hook,
            dir = dir,
            is_permanent = is_permanent,
            prepend = prepend,
        )
        hook_point.add_hook(hook, dir=dir, is_permanent = is_permanent, level = level, prepend = prepend)
    
    def check_hooks_to_add(
        self, hook_point, hook_point_name, hook, dir="fwd", is_permanent=False, prepend=False) -> None:
        """Override this function to add checks on which hooks should be added (ex: see base_vit.py)"""
        pass 
    
    def add_hook(self, name, hook, dir="fwd", is_permanent=False, level=None, prepend=False) -> None:
        
        if type(name) == str:
            self.check_and_add_hook(
                self.mod_dict[name],
                name,
                hook,
                dir = dir,
                is_permanent = is_permanent,
                level = level,
                prepend = prepend,
            )
        else:
            for hook_point_name, hp in self.hook_dict.items():
                if name(hook_point_name):
                    self.check_and_add_hook(
                        hp,
                        hook_point_name,
                        hook,
                        dir = dir,
                        is_permanent = is_permanent,
                        level = level,
                        prepend = prepend,
                    )
    
    def add_perma_hook(self, name, hook, dir="fwd") -> None:
        self.add_hook(name, hook, dir, is_permanent=True)
    
    @contextmanager
    def hooks(
        self,
        fwd_hooks: List[Tuple[Union[str, Callable], Callable]] = [],
        bwd_hooks: List[Tuple[Union[str, Callable], Callable]] = [],
        reset_hooks_end: bool = True,
        clear_contexts: bool = True
    ):
        try:
            self.context_level += 1

            for name, hook in fwd_hooks:
                if type(name) == str:
                    self.mod_dict[name].add_hook(
                        hook, dir="fwd", level=self.context_level
                    )
                else:
                    for hook_point_name, hp in self.hook_dict.items():
                        if name(hook_point_name):
                            hp.add_hook(hook, dir="fwd", level=self.context_level)

            for name, hook in bwd_hooks:
                if type(name) == str:
                    self.mod_dict[name].add_hook(
                        hook, dir="bwd", level=self.context_level
                    )
                else:
                    for hook_point_name, hp in self.hook_dict.items():
                        if name(hook_point_name):
                            hp.add_hook(hook, dir="bwd", level=self.context_level)
            yield self
        finally:
            if reset_hooks_end:
                self.reset_hooks(clear_contexts=clear_contexts, including_permanent=False, level=self.context_level)
            self.context_level -= 1
    
    def run_with_hooks(
        self,
        *model_args,
        fwd_hooks: List[Tuple[Union[str, Callable], Callable]] = [],
        bwd_hooks: List[Tuple[Union[str, Callable], Callable]] = [],
        reset_hooks_end = True,
        clear_contexts = False,
    ):
        if len(bwd_hooks) > 0 and reset_hooks_end:
            logging.warning(
                "WARNING: Hooks will be reset at the end of run_with_hooks. This removes the backward hooks before a backward pass can occur."
            )
        with self.hooks(
            fwd_hooks, bwd_hooks,reset_hooks_end, clear_contexts
        ) as hooked_model:
            return hooked_model.forward(*model_args)

    def add_caching_hooks(
            self,
            names_filter: NamesFilter = None,
            incl_bwd: bool = False,
            device = None,
            remove_batch_dim: bool = False,
            cache: Optional[dict] = None,
    ) -> dict:
        
        if cache is None:
            cache = {}
        
        if names_filter is None:
            names_filter = lambda name: True
        elif type(names_filter) == str:
            filter_str = names_filter
            names_filter = lambda name: name == filter_str
        elif type(names_filter) == list:
            filter_list = names_filter
            names_filter = lambda name: name in filter_list

        self.is_caching = True

        def _save_hook(tensor, hook):
            if remove_batch_dim:
                cache[hook.name] = tensor.detach().to(device)[0]
            else:
                cache[hook.name] = tensor.detach().to(device)
        
        def _save_hook_back(tensor, hook):
            if remove_batch_dim:
                cache[hook.name + "_grad"] = tensor.detach().to(device)[0]
            else:
                cache[hook.name + "_grad"] = tensor.detach().to(device)
        
        for name, hp in self.hook_dict.items():
            if names_filter(name):
                hp.add_hook(_save_hook, dir="fwd")
                if incl_bwd:
                    hp.add_hook(_save_hook_back, dir="bwd")

        return cache

    def run_with_cache(
            self,
            *model_args,
            names_filter: NamesFilter = None,
            device = None,
            remove_batch_dim: bool = False,
            incl_bwd = False,
            reset_hooks_end = True,
            clear_contexts = False,
            **model_kwargs,

    ):
        cache_dict, fwd, bwd = self.get_caching_hooks(
            names_filter, incl_bwd, device, remove_batch_dim=remove_batch_dim
        )
        with self.hooks(fwd_hooks=fwd, bwd_hooks=bwd, reset_hooks_end=reset_hooks_end, clear_contexts=clear_contexts):
            model_out = self(*model_args, **model_kwargs)
            if incl_bwd:
                model_out.backward()
        return model_out, cache_dict

    def get_caching_hooks(
            self,
            names_filter: NamesFilter = None,
            incl_bwd: bool = False,
            device = None,
            remove_batch_dim: bool = False,
            cache: Optional[dict] = None,
    ) -> Tuple[dict, list, list]:
        
        if cache is None:
            cache = {}
        
        if names_filter is None:
            names_filter = lambda name: True
        elif type(names_filter) == str:
            filter_str = names_filter
            names_filter = lambda name: name == filter_str
        elif type(names_filter) == list:
            filter_list = names_filter
            names_filter = lambda name: name in filter_list
        
        self.is_caching = True

        def _save_hook(tensor, hook):
            if remove_batch_dim:
                cache[hook.name] = tensor.detach().to(device)[0]
            else:
                cache[hook.name] = tensor.detach().to(device)
        
        def _save_hook_back(tensor, hook):
            if remove_batch_dim:
                cache[hook.name + "_grad"] = tensor.detach().to(device)[0]
            else:
                cache[hook.name + "_grad"] = tensor.detach().to(device)
        
        fwd_hooks = []
        bwd_hooks = []
        for name, hp in self.hook_dict.items():
            if names_filter(name):
                fwd_hooks.append((name, _save_hook))
                if incl_bwd:
                    bwd_hooks.append((name, _save_hook_back))
        
        return cache, fwd_hooks, bwd_hooks



    




import random
import numpy as np
import os
from sklearn.model_selection import train_test_split
from torch.utils.data import Dataset
import torch

class InductionDataset(Dataset):
    def __init__(self, train_or_test, dir_path = '../data/induction', use_metadata=False, transform=None):

        self.dir_path = dir_path
        self.cache_path = f'{dir_path}/all_{train_or_test}.npz'

        self.use_metadata = use_metadata
        self.transform = transform

        if not os.path.exists(self.cache_path):
            print("Generating and saving new induction dataset...")
            self._generate_and_cache()

        print("Loading induction dataset from cache...", self.cache_path)
        self._load_from_cache()
        # self._normalize_and_to_tensor()

    def __len__(self):
        return len(self.images)

    def __getitem__(self, idx):
        image = self.images[idx][np.newaxis, :, :]
        label = self.labels[idx]
        # meta_data = self.metadata[idx]
        image = torch.from_numpy(image).float()
        if self.transform:
            image = self.transform(image)
        return image, label

    def _load_from_cache(self):
        loaded = np.load(self.cache_path, allow_pickle=True, mmap_mode='r' )
        self.images = loaded['images']
        if self.use_metadata:
            self.metadata = loaded['metadata']
        self.labels = loaded['labels']

    def _generate_and_cache(self):
        generate_dataset(self.dir_path)

    # def _normalize_and_to_tensor(self):
    #     self.images = 2 * torch.tensor(self.images, dtype=torch.float32) / 255.0 - 1 # normalize [-1,]
    #     self.labels = torch.tensor(self.labels, dtype=torch.int64)

## Functionality to generate dataset ##

def draw_circle(image, center_row, center_col, radius=2, im_size=32):
        """Draw a circle on the given image."""
        for r in range(center_row-radius, center_row+radius+1):
            for c in range(center_col-radius, center_col+radius+1):
                if (r - center_row)**2 + (c - center_col)**2 <= radius**2 and 0 <= r < im_size and 0 <= c < im_size:
                    image[r, c] = 1
        return image

def draw_line(image, center_row, center_col, line_length=4, im_size=32):
    for i in range(-line_length // 2, line_length // 2 + 1):
        if 0 <= center_row + i < im_size and 0 <= center_col < im_size:
            image[center_row + i, center_col] = 1
    return image

def draw_x(image, center_row, center_col, x_length=5, im_size=32):
    # Drawing the X centered around the center_col
# Drawing the X centered around the start_col
    for i in range(x_length):
        image[center_row - x_length // 2 + i, center_col - x_length // 2 + i] = 1
        image[center_row - x_length // 2 + i, center_col + x_length // 2 - i] = 1
    return image


def draw_diagonal(image, center_row, center_col, line_length=4, im_size=32):
    for i in range(-line_length // 2, line_length // 2 + 1):
        if 0 <= center_row + i < im_size and 0 <= center_col + i < im_size:
            image[center_row + i, center_col + i] = 1

    return image

def plot_two_objects(A, B, Ax, Ay, Bx, By, vertical=False):

    image = np.zeros((32, 32))

    # List of available drawing functions
    draw_functions = [draw_circle, draw_line, draw_x, draw_diagonal]

    # Call the first chosen function
    A(image, Ax, Ay)

    # Call the second chosen function right next to the first
    B(image, Bx, By)

    if vertical:
        image = image.T

    return image

def generate_dataset(dir_path = '../data/induction'):

    # generate one of each image combo, make sure spacing makes sense
    draw_functions = [draw_circle, draw_line, draw_x, draw_diagonal]
    padding = 4
    offset = 7

    images = []
    metadata = []
    labels = [] # from 0 to 3

    for vertical in [True, False]:
        for a in range(padding, 32 - padding):
            for b in range(padding, 32 - padding - offset):
                Ax = a
                Ay = b
                Bx = Ax
                By = Ay + offset

                # Example of how to use it:
                for A in draw_functions:
                    for B in draw_functions:
                        img = plot_two_objects(A, B, Ax, Ay, Bx, By, vertical=vertical)

                        if A == B:
                            same = True
                        else:
                             same = False

                        images.append(img)
                        m = {
                            "Ax": Ax,
                            "Ay": Ay,
                            "Bx": Bx,
                            "By": By,
                            "A": A.__name__,
                            "B": B.__name__,
                            "Same": same,
                            "Vertical": vertical
                        }
                        metadata.append(m)

                        if vertical and same:
                            l = 0
                        elif vertical and not same:
                            l = 1
                        elif not vertical and same:
                            l = 2
                        elif not vertical and not same:
                            l = 3

                        labels.append(l)
    
    path = f'{dir_path}/induction_dataset.npz'
    print(f"Saving raw dataset to {path}...")
    if not os.path.exists(dir_path):
        os.makedirs(dir_path)
    np.savez(path, images=images, metadata=metadata, labels=labels)

    print("Balancing dataset...")
    create_balanced_dataset(dir_path)


def create_balanced_dataset(dir_path='../data/induction'):
    
    # Load the dataset
    dataset = np.load(f'{dir_path}/induction_dataset.npz', mmap_mode='r', allow_pickle=True)
    images, metadata, labels = dataset['images'], dataset['metadata'], dataset['labels']

    # Create categories based on the conditions
    categories = {
        'same_T_vertical_T': [i for i, entry in enumerate(metadata) if entry['Same'] and entry['Vertical']],
        'same_F_vertical_F': [i for i, entry in enumerate(metadata) if not entry['Same'] and not entry['Vertical']],
        'same_T_vertical_F': [i for i, entry in enumerate(metadata) if entry['Same'] and not entry['Vertical']],
        'same_F_vertical_T': [i for i, entry in enumerate(metadata) if not entry['Same'] and entry['Vertical']]
    }

    sample_size = min(len(c) for c in categories.values())

    # Sample from each category and save as separate npz files
    for key, val in categories.items():
        indices = random.sample(val, sample_size)
        np.savez(f'{dir_path}/{key}.npz', 
                 images=images[indices], 
                 metadata=metadata[indices], 
                 labels=labels[indices])

    # Split each npz file into train/test datasets
    all_train_data, all_test_data = [], []
    for key in categories.keys():
        data = np.load(f'{dir_path}/{key}.npz', allow_pickle=True)
        train, test = train_test_split(list(zip(data['images'], data['metadata'], data['labels'])), test_size=0.1, random_state=42)
        all_train_data.extend(train)
        all_test_data.extend(test)

    # Shuffle and save combined train/test datasets
    random.shuffle(all_train_data)
    random.shuffle(all_test_data)

    train_images, train_metadata, train_labels = zip(*all_train_data)
    test_images, test_metadata, test_labels = zip(*all_test_data)

    np.savez(f'{dir_path}/all_train.npz', images=np.array(train_images), metadata=np.array(train_metadata), labels=np.array(train_labels))
    np.savez(f'{dir_path}/all_test.npz', images=np.array(test_images), metadata=np.array(test_metadata), labels=np.array(test_labels))
    print("Saved total train and test.")



"""
Prisma Repo
By Sonia Joseph

Copyright (c) Sonia Joseph. All rights reserved.

Inspired by TransformerLens. Some functions have been adapted from the TransformerLens project.
For more information on TransformerLens, visit: https://github.com/neelnanda-io/TransformerLens
"""

from dataclasses import dataclass

from typing import Optional

import torch.utils.hooks as hooks


@dataclass
class LensHandle:
    """Dataclass that holds information about a PyTorch hook."""

    hook: hooks.RemovableHandle
    """Reference to the Hook's Removable Handle."""

    is_permanent: bool = False
    """Indicates if the Hook is Permanent."""

    context_level: Optional[int] = None
    """Context level associated with the hooks context manager for the given hook."""





import pytest
import torch
from vit_prisma.models.base_vit import HookedViT
import numpy as np
import torch

from transformers import VivitForVideoClassification



def test_loading_vivet():
    TOLERANCE = 1e-4

    model_name = "google/vivit-b-16x2-kinetics400"
    batch_size = 5
    channels = 3
    num_frames = 32
    height = 224
    width = 224
    device = "cpu"

    hooked_model = HookedViT.from_pretrained(model_name, is_timm=False)
    hooked_model.to(device)
    google_model = VivitForVideoClassification.from_pretrained(model_name)
    google_model.to(device)

    with torch.random.fork_rng():
        torch.manual_seed(1)
        input_image = torch.rand((batch_size, num_frames, channels, height, width)).to(device)
    with torch.no_grad():
        hooked_output, timm_output = hooked_model(input_image), google_model(input_image).logits

    assert torch.allclose(hooked_output, timm_output, atol=TOLERANCE), f"Model output diverges! Max diff: {torch.max(torch.abs(hooked_output - timm_output))}"





import torch
import torch.nn as nn

import torch
from torch.utils.data import Dataset, DataLoader, random_split
from itertools import combinations
import torchvision.transforms as transforms

import numpy as np

from PIL import Image, ImageDraw
import math

import random

import os

from torchvision.transforms.transforms import Grayscale, ToTensor

def check_if_valid_angle(angle, angle_range):
    if angle < 0 or angle > 360:
        raise ValueError('Angle must be between 0 and 360')
    
    # Check if in range
    if not np.isin(angle, angle_range):
        raise ValueError('Angle must at correct interval of angle_range')
    
# def get_datasets(split_ratio, model_type):
#     circle_metadata = get_circle_metadata()
#     # Create train and test data once
#     train_data_raw, test_data_raw = get_train_test_data(circle_metadata, split_ratio) 
#     train_dataset = CircleDataset(circle_metadata, train_data_raw, train_or_test='train', model_type = model_type)
#     test_dataset = CircleDataset(circle_metadata, test_data_raw, train_or_test='test', model_type = model_type)
#     return train_dataset, test_dataset

def get_train_test_data(circle_metadata, split_ratio=0.5):
    data = list(combinations(range(0, circle_metadata['mod_arith']), 2))
    random.shuffle(data)
    split_idx = int(len(data) * split_ratio)
    return data[:split_idx], data[split_idx:]
    
def get_circle_metadata():
    circle_metadata = {
    "mod_arith": 60,
    "image_size": 32,
    "center": (16, 16),
    "radius": 15.5,
    "multiplier": 6,
    "angle_range": np.arange(0, 60, 1)
    } 
    return circle_metadata

def draw_circle_with_points(angle1=None, angle2=None, metadata=None, model_type=None):

    # Create a new image with white background
    image_size = metadata['image_size']
    img = Image.new('L', (image_size, image_size), color=255) # white image

    pixels = img.load()

    center = metadata['center']
    radius = metadata['radius']

    # Transform angle to get point in circle
    MULTIPLIER = metadata['multiplier']
    angle_range = metadata['angle_range']
    transformed_angle_range = angle_range * MULTIPLIER

    # Draw circle in black.
    for i in transformed_angle_range:
        x = center[0] + radius * math.cos(math.radians(i))
        y = center[1] + radius * math.sin(math.radians(i))
        pixels[x, y] = 0

    # Specify the angle
    def _draw_point_(angle, color=128):
        angle_rad = math.radians(angle) # x3 b/c representing the circle in intervals of 3
        x = center[0] + radius * math.cos(angle_rad)
        y = center[1] + radius * math.sin(angle_rad)
        pixels[x, y] = color

    if angle1 is not None:
        check_if_valid_angle(angle1, angle_range)
        _draw_point_(angle1*MULTIPLIER)
    if angle2 is not None:
        check_if_valid_angle(angle2, angle_range)
        _draw_point_(angle2*MULTIPLIER)


    if model_type == 'pretrained_transformer':

        # Define padding dimensions
        left_padding = (224 - 32) // 2
        top_padding = (224 - 32) // 2
        right_padding = 224 - 32 - left_padding
        bottom_padding = 224 - 32 - top_padding

        transform = transforms.Compose([
        transforms.Pad((left_padding, top_padding, right_padding, bottom_padding), fill=1),  # Assuming white padding,  # Resize the image to 224x224
        transforms.Grayscale(num_output_channels=3),  
        transforms.ToTensor()           # Convert the PIL Image to a tensor
    ])
    else:
        transform = transforms.Compose([
        transforms.Grayscale(num_output_channels=1),
        transforms.ToTensor(),
        transforms.Normalize([0.5], [0.5])  # Normalize to [-1, 1]
    ])

    img = transform(img)
    
    return img 

class CircleDataset(Dataset):
    def __init__(self, train_or_test='test', 
    cache_path = '/home/mila/s/sonia.joseph/ViT-Planetarium/data/circle',
    transform=None):

        self.cache_path = cache_path
        self.train_or_test = train_or_test

        # self.circle_metadata = circle_metadata
        # self.mod_arith = circle_metadata['mod_arith']
        # self.data = [{'data': item, 'metadata': i} for i, item in enumerate(data)]
        # self.model_type = model_type

        self.transform=transform

        if os.path.exists(f'{cache_path}/{train_or_test}.npz'):
            print("Loading circle dataset from cache...", f'{cache_path}/{train_or_test}.npz')
            self._load_from_cache()
        else:
            print("Generating and saving new circle dataset...")
            self._generate_and_cache()

    def _load_from_cache(self):
        loaded = np.load(f'{self.cache_path}/{self.train_or_test}.npz', allow_pickle=True)
        self.imgs = loaded['imgs']
        self.labels = loaded['labels']
        self.data_points = loaded['data_points']

    def __len__(self):
        return len(self.imgs)

    def _generate_and_cache(self):
        imgs = []
        labels = []
        data_points = []
        
        for idx in range(len(self.data)):
            data_point = self.data[idx]['data']
            label = sum(data_point) % self.mod_arith
            img = draw_circle_with_points(data_point[0], data_point[1], self.circle_metadata, model_type=self.model_type)
            
            imgs.append(img)
            labels.append(label)
            data_points.append(data_point)

        self.imgs = np.array(imgs)
        self.labels = np.array(labels)
        self.data_points = np.array(data_points)

        # Cache the dataset
        np.savez(f'{self.cache_path}/{self.train_or_test}.npz', imgs=self.imgs, labels=self.labels, data_points=self.data_points)

    def __getitem__(self, idx):
        image = self.imgs[idx]
        label = self.labels[idx]
        # meta_data = self.metadata[idx]
        if self.transform:
            image = self.transform(image)
        return image, label




# Helper function: javascript visualization
import numpy as np
import json
from IPython.core.display import display, HTML
import string
import random

# Helper function to plot attention patterns, hide this function

import matplotlib.pyplot as plt
import numpy as np

def plot_attn_heads(total_activations, n_heads=12, n_layers=12, img_shape=50, idx=0, figsize=(20, 20),
                    global_min_max=False, global_normalize=False, fourier_transform_local=False, log_transform=False,
                    fourier_transform_global=False, graph_type="imshow_graph", cmap='viridis'):

    # New shape handling: total_activations is now expected to be of shape [n_layers*n_heads, img_shape, img_shape]
    total_data = np.zeros((n_layers*n_heads, img_shape, img_shape))

    # Adjusted processing for flattened layer-heads structure
    if global_min_max or global_normalize or fourier_transform_global:
        for i in range(n_layers*n_heads):
            data = total_activations[i, :, :]
            if log_transform:
                data = np.log10(np.maximum(data, 1e-6))  # log10_stable equivalent
            if fourier_transform_global:
                data = np.abs(np.fft.fftshift(np.fft.fft2(data)))
            total_data[i, :, :] = data

        total_min, total_max = np.min(total_data), np.max(total_data)
        print(f"Total Min: {total_min}, Total Max: {total_max}")

        if global_normalize:
            total_data = -1 + 2 * (total_data - total_min) / (total_max - total_min)

    fig, axes = plt.subplots(n_layers, n_heads, figsize=figsize, squeeze=False)  # Ensure axes is always 2D array
    total_data_dict = {}

    for i in range(n_layers):
        total_data_dict[f"Layer_{i}"] = {}
        for j in range(n_heads):
            # Adjust indexing for the flattened layer-head structure
            linear_idx = i * n_heads + j
            data = total_data[linear_idx, :, :]

            if graph_type == "histogram_graph":
                data = data.flatten()
                axes[i, j].hist(data, bins=100, log=log_transform, cmap=cmap)
            elif graph_type == "imshow_graph":
                if fourier_transform_local:
                    data = np.abs(np.fft.fftshift(np.fft.fft2(data)))
                vmin, vmax = (total_min, total_max) if (global_min_max or global_normalize) else (data.min(), data.max())
                im = axes[i, j].imshow(data, vmin=vmin, vmax=vmax, cmap=cmap)
                axes[i, j].axis('off')
                total_data_dict[f"Layer_{i}"][f"Head_{j}"] = data.tolist()

            axes[i, j].set_title(f"Head {j}", fontsize=12, pad=5) if i == 0 else None
            if j == 0:
                axes[i, j].text(-0.3, 0.5, f"Layer {i}", fontsize=12, rotation=90, ha='center', va='center', transform=axes[i, j].transAxes)

    # Add colorbar for imshow_graph
    if graph_type == "imshow_graph" and (global_min_max or global_normalize):
        cbar_ax = fig.add_axes([0.92, 0.15, 0.02, 0.7])
        fig.colorbar(im, cax=cbar_ax)
        cbar_ax.set_title('Attention', size=12)

    plt.subplots_adjust(wspace=0.2, hspace=0.4)
    plt.suptitle(f'Attention for Image Idx {idx}', fontsize=20, y=0.93)
    plt.show()




from setuptools import setup, find_packages


install_requires = [
    'pytest>=6.0',  # Test dependency, but included here to auto-install
    'torch',        # Test dependency
    'numpy',        # Test dependency
    'jaxtyping',    # Test dependency
    'einops',       # Test dependency
    'fancy_einsum', # Test dependency
    'plotly==5.19.0',
    'timm',         # Test dependency
    'transformers', # Test dependency
    'scikit-learn', # Test dependency
    'datasets',
]

setup(
    name='vit-prisma',
    version='0.1.4',
    author='Sonia Joseph',
    author_email='soniamollyjoseph@gmail.com',
    description='A Vision Transformer library for mechanistic interpretability.',
    long_description=open('docs/README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/soniajoseph/vit-prisma',
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    package_data={
    'vit_prisma': ['visualization/*.html', 'visualization/*.js'],
    # Add other patterns here as needed
},
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    python_requires='>=3.6',
    install_requires=install_requires,
    keywords='vision-transformer, clip, multimodal, machine-learning, mechanistic interpretability',
    zip_safe=False,
)



# import unittest
# from torchvision import datasets, transforms
# from torch.nn import CrossEntropyLoss
# from vit_prisma.models.base_vit import BaseViT  # Assuming you have a ViT model defined somewhere

# from vit_prisma.configs.MNISTConfig import GlobalConfig
# from vit_prisma.training.trainer import train

# class TestTrainingFunction(unittest.TestCase):

#         # Load MNIST dataset
#         def setUp(self):

#             data_root_dir = '../../data'
#             transform = transforms.Compose([transforms.ToTensor(), transforms.Normalize((0.5,), (0.5,))])
#             self.train_dataset = datasets.MNIST(root=data_root_dir, train=True, transform=transform, download=True)
#             self.val_dataset = datasets.MNIST(root=data_root_dir, train=False, transform=transform, download=True)

#             self.config = GlobalConfig()

#             # Define a simple model
#             self.model_function = BaseViT

#         def test_train_function(self):
#             self.config.logging.wandb_team_name = 'perceptual-alignment'
#             self.config.training.early_stopping = False

#             trained_model = train(self.model_function, self.config, self.train_dataset, self.val_dataset)
#             self.assertIsInstance(trained_model, type(self.model))

# if __name__ == '__main__':
#     unittest.main()



import logging
import torch.nn as nn
import torch 
from jaxtyping import Float
from vit_prisma.configs import HookedViTConfig
import einops

class PatchEmbedding(nn.Module):

    def __init__(self, config, logger=None):
        super().__init__()
        self.logger = logger
        self.config = config
        self.proj = nn.Conv2d(
            self.config.n_channels, 
            self.config.d_model, 
            kernel_size=self.config.patch_size, 
            stride=self.config.patch_size, 
            bias=True
        )
    
    def _log(self, stage, tensor):
        if self.logger:
            self.logger.info(f"{stage} size: {tensor.shape}")

    def forward(self, x:Float[torch.Tensor, "batch channel height width"]) -> Float[torch.Tensor, "batch n_tokens d_model"]:
        self._log("PatchEmbedding input", x)

        x = self.proj(x).flatten(2).transpose(1, 2)

        self._log("PatchEmbedding output", x)
        return x
    

# Used for videos, 3 dimensional spacetime patches 
class TubeletEmbedding(nn.Module):

    def __init__(self, cfg:HookedViTConfig):
        super().__init__()
        self.cfg = cfg

        tubelet_size = [self.cfg.video_tubelet_depth, self.cfg.patch_size, self.cfg.patch_size]
        self.proj = nn.Conv3d(
            self.cfg.n_channels, 
            self.cfg.d_model, 
            kernel_size=tubelet_size, 
            stride=tubelet_size, 
            bias=True
        )

    

    def forward(self, x:Float[torch.Tensor, "batch num_frames channels height width"]) -> Float[torch.Tensor, "batch n_tokens d_model"]:
        
        # Flip num_frames and channels
        x = einops.rearrange(x, "b t c h w -> b c t h w")
        
        x = self.proj(x)

        # Flatten the tokens
        x = einops.rearrange(x, "b c t h w -> b (t h w) c") 

        return x




"""
Prisma Repo
By Sonia Joseph

Copyright (c) Sonia Joseph. All rights reserved.

Inspired by TransformerLens. Some functions have been adapted from the TransformerLens project.
For more information on TransformerLens, visit: https://github.com/neelnanda-io/TransformerLens
"""

import numpy as np
import torch
from collections import defaultdict

from vit_prisma.utils.data_utils.imagenet_dict import IMAGENET_DICT
from vit_prisma.utils.data_utils.imagenet_utils import imagenet_index_from_word


def get_patch_logit_directions(cache, all_answers, incl_mid=False, return_labels=True):
    accumulated_residual, labels = cache.accumulated_resid(
        layer=-1, incl_mid=incl_mid, return_labels=True
    )
    scaled_residual_stack = cache.apply_ln_to_stack(
        accumulated_residual, layer=-1,
    )
    result = torch.einsum('lbpd,od -> lbpo', scaled_residual_stack, all_answers)
    # Rearrange so batches are first
    result = result.permute(1, 2, 0, 3)
    return result, labels

def get_patch_logit_dictionary(patch_logit_directions, batch_idx=0, rank_label=None):
    patch_dictionary = defaultdict(list)
    # if tuple, get first entry
    if isinstance(patch_logit_directions, tuple):
        patch_logit_directions = patch_logit_directions[0]
    # Go through laeyrs of one batch
    for patch_idx, patches in enumerate(patch_logit_directions[batch_idx]):
        # Go through every patch and get max prediction
        for logits in patches:
            probs = torch.softmax(logits, dim=-1)
            # Get index of max prediction
            predicted_idx = int(torch.argmax(probs))
            logit = logits[predicted_idx].item()
            predicted_class_name = IMAGENET_DICT[predicted_idx]
            if rank_label:
                # Where is the rank_label in the sorted list?
                rank_index = imagenet_index_from_word(rank_label)
                sorted_list = torch.argsort(probs, descending=True)
                rank = np.where(sorted_list == rank_index)[0][0]
                patch_dictionary[patch_idx].append((logit, predicted_class_name, predicted_idx, rank))
            else:
                patch_dictionary[patch_idx].append((logit, predicted_class_name, predicted_idx))
    return patch_dictionary



"""
Prisma Repo
By Sonia Joseph

Copyright (c) Sonia Joseph. All rights reserved.

Inspired by TransformerLens. Some functions have been adapted from the TransformerLens project.
For more information on TransformerLens, visit: https://github.com/neelnanda-io/TransformerLens
"""

from typing import List, Union, Dict, Callable, Tuple, Optional, Any
from vit_prisma.prisma_tools.lens_handle import LensHandle
import torch.nn as nn


class HookPoint(nn.Module):
    """
    A helper class to access intermediate activations in a PyTorch model (adapted from TransformerLens, which was inspired by Garcon).

    HookPoint is a dummy module that acts as an identity function by default. By wrapping any
    intermediate activation in a HookPoint, it provides a convenient way to add PyTorch hooks.
    """
    def __init__(self):
        super().__init__()
        self.fwd_hooks: List[LensHandle] = []
        self.bwd_hooks: List[LensHandle] = []

        self.ctx = {} # what is this?

        self.name = None
    
    def add_perma_hook(self, hook, dir="fwd") -> None:
        self.add_hook(hook, dir, is_permanent=True)
    
    def add_hook(
            self, hook, dir="fwd", is_permanent=False, level=None, prepend=False
    ) -> None:
        """
        If prepend is True, add this hook before all other hooks.
        """

        if dir == "fwd":

            def full_hook(module, module_input, module_output):
                return hook(module_output, hook=self)

            full_hook.__name__ = (
                hook.__repr__()
            )

            handle = self.register_forward_hook(full_hook)
            handle = LensHandle(handle, is_permanent, level)

            if prepend:
                self._forward_hooks.move_to_end(handle.hook.id, last=False)
                self.fwd_hooks.insert(0, handle)
            else:
                self.fwd_hooks.append(handle)

        elif dir == "bwd":

            def full_hook(module, module_input, module_output):
                return hook(module_output[0], hook=self)
            
            full_hook.__name__ = (
                hook.__repr__()
            )

            handle = self.register_backward_hook(full_hook)
            handle = LensHandle(handle, is_permanent, level)
        
            if prepend:
                self._backward_hooks.move_to_end(handle.hook.id, last=False)
                self.bwd_hooks.insert(0, handle)
            else:
                self.bwd_hooks.append(handle)
        
        else :
            raise ValueError(f"Invalid dir {dir}. dir must be 'fwd' or 'bwd'")

    def remove_hooks(self, dir="fwd", including_permanent=False, level=None) -> None:
        def _remove_hooks(handles: List[LensHandle]) -> List[LensHandle]:
            output_handles = []
            for handle in handles:
                if including_permanent:
                    handle.hook.remove()
                elif (not handle.is_permanent) and (level is None or handle.context_level == level):
                    handle.hook.remove()
                else:
                    output_handles.append(handle)
            return output_handles
        
        if dir == "fwd" or dir == "both":
            self.fwd_hooks = _remove_hooks(self.fwd_hooks)
        elif dir == "bwd" or dir == "both":
            self.bwd_hooks = _remove_hooks(self.bwd_hooks)
        else:
            raise ValueError(f"Invalid direction {dir}. dir must be 'fwd', 'bwd', or 'both'")

    def clear_context(self):
        del self.ctx
        self.ctx = {}

    def forward(self, x):
        return x

    def layer(self):
        # Returns the layer index if the name has the form 'blocks.{layer}.{...}'
        # Helper function that's mainly useful on HookedTransformer
        # If it doesn't have this form, raises an error -
        split_name = self.name.split(".")
        return int(split_name[1])





from vit_prisma.utils.data_utils.imagenet_dict import IMAGENET_DICT

def imagenet_index_from_word(search_term):
    # Convert the search term to lowercase to ensure case-insensitive matching
    search_term = search_term.lower()

    # Iterate over the dictionary and search for the term
    for key, value in IMAGENET_DICT.items():
        if search_term in value.lower():  # Convert each value to lowercase for case-insensitive comparison
            return key  # Return the key directly once found

    # If the loop completes without returning, the term was not found; raise an exception
    raise ValueError(f"'{search_term}' not found in IMAGENET_DICT.")


import torch

activation_dict ={
    "ReLU": torch.nn.ReLU,
    "LeakyReLU": torch.nn.LeakyReLU,
    "GELU": torch.nn.GELU,
    "Linear": torch.nn.Identity,
}

initialization_dict = {
    "ReLU": 'relu',
    "LeakyReLU": 'relu',
    "GELU": 'relu', # for 'he' intitializatoin
}

optimizer_dict = {
    "AdamW": torch.optim.AdamW,
}

loss_function_dict = {
    "CrossEntropy": torch.nn.CrossEntropyLoss,
    "MSE": torch.nn.MSELoss,
}


import torch
import torch.nn as nn

from vit_prisma.models.layers.attention import Attention
from vit_prisma.models.layers.mlp import MLP
from vit_prisma.models.layers.layer_norm import LayerNorm, LayerNormPre

from vit_prisma.configs.HookedViTConfig import HookedViTConfig
from vit_prisma.prisma_tools.hook_point import HookPoint

from typing import Dict, Optional, Tuple, Union

from jaxtyping import Float, Int
import einops



def add_head_dimension(
        tensor: Float[torch.Tensor, "batch pos d_model"],
        n_heads:int,
        clone_tensor=True,
        # `einops.repeat` uses a view in torch, so we generally clone the tensor to avoid using shared storage for each head entry
    ):
        repeated_tensor = einops.repeat(
            tensor,
            "batch pos d_model -> batch pos n_heads d_model",
            n_heads=n_heads,
        )
        if clone_tensor:
            return repeated_tensor.clone()
        else:
                return repeated_tensor

class TransformerBlock(nn.Module):
    """
    Transformer block.
    """

    def __init__(self, cfg: Union[Dict, HookedViTConfig], block_index=None):
        super().__init__()

        if isinstance(cfg, Dict):
            cfg = HookedViTConfig.from_dict(cfg)
        
        self.cfg = cfg

        if self.cfg.normalization_type == "LN":
            self.ln1 = LayerNorm(self.cfg)
            if not self.cfg.attn_only:
                self.ln2 = LayerNorm(self.cfg)
        elif self.cfg.normalization_type == "LNPre":
            self.ln1 = LayerNormPre(self.cfg)
            if not self.cfg.attn_only:
                self.ln2 = LayerNormPre(self.cfg)
        elif self.cfg.normalization_type is None:
            self.ln1 = nn.Identity()
            if not self.cfg.attn_only:
                self.ln2 = nn.Identity()
        else:
            raise ValueError(f"Invalid normalization type: {self.cfg.normalization_type}")
        
        self.attn  = Attention(self.cfg)

        if not self.cfg.attn_only:
            self.mlp = MLP(self.cfg)
        
        self.hook_attn_in = HookPoint()
        self.hook_q_input = HookPoint()
        self.hook_k_input = HookPoint()
        self.hook_v_input = HookPoint()
        self.hook_mlp_in = HookPoint()
        self.hook_attn_out = HookPoint()
        self.hook_mlp_out = HookPoint()
        self.hook_resid_pre = HookPoint()

        if not self.cfg.attn_only:
            self.hook_resid_mid = HookPoint()

        self.hook_resid_post = HookPoint()

    def forward(
            self,
            resid_pre: Float[torch.Tensor, "batch pos d_model"],
    ) -> Float[torch.Tensor, "batch pos d_model"]:
        
        resid_pre = self.hook_resid_pre(resid_pre)

        if self.cfg.use_attn_in or self.cfg.use_split_qkv_input:
            # We're adding a head dimension
            attn_in = add_head_dimension(resid_pre, self.cfg.n_heads, clone_tensor=False)
        else:
            attn_in = resid_pre

        if self.cfg.use_attn_in:
            attn_in = self.hook_attn_in(attn_in.clone())
        
        if self.cfg.use_split_qkv_input:
            query_input = self.hook_q_input(attn_in.clone())
            key_input = self.hook_k_input(attn_in.clone())
            value_input = self.hook_v_input(attn_in.clone())
        else:
            query_input = attn_in
            key_input = attn_in
            value_input = attn_in
        
        attn_out = self.attn(
                query_input = self.ln1(query_input),
                key_input = self.ln1(key_input),
                value_input = self.ln1(value_input),
            )

        # Take hook fn
        
        attn_out = self.hook_attn_out(
            attn_out
        )

        if not self.cfg.attn_only: 
            resid_mid = self.hook_resid_mid(
                resid_pre + attn_out
            )
            mlp_in = (
                resid_mid
                if not self.cfg.use_hook_mlp_in
                else self.hook_mlp_in(resid_mid.clone())
            )
            normalized_resid_mid = self.ln2(mlp_in)
            mlp_out = self.hook_mlp_out(self.mlp(normalized_resid_mid))
            resid_post = self.hook_resid_post(resid_mid + mlp_out)
        else:
            resid_post = self.hook_resid_post(resid_pre + attn_out)
        
        return resid_post

    
class BertBlock(nn.Module):
    """
    
    Just like TransformerBlock but applies layernorm after attn and after mlp, not before. Necessary for some CLIP models.
    """

    def __init__(self, cfg: Union[Dict, HookedViTConfig], block_index):
        super().__init__()

        if isinstance(cfg, Dict):
            cfg = HookedViTConfig.from_dict(cfg)
        
        self.cfg = cfg

        if self.cfg.normalization_type == "LN":
            self.ln1 = LayerNorm(self.cfg)
            if not self.cfg.attn_only:
                self.ln2 = LayerNorm(self.cfg)
        elif self.cfg.normalization_type == "LNPre":
            self.ln1 = LayerNormPre(self.cfg)
            if not self.cfg.attn_only:
                self.ln2 = LayerNormPre(self.cfg)
        elif self.cfg.normalization_type is None:
            self.ln1 = nn.Identity()
            if not self.cfg.attn_only:
                self.ln2 = nn.Identity()
        else:
            raise ValueError(f"Invalid normalization type: {self.cfg.normalization_type}")
        
        self.attn  = Attention(self.cfg)

        if not self.cfg.attn_only:
            self.mlp = MLP(self.cfg)
        
        self.hook_attn_in = HookPoint()
        self.hook_q_input = HookPoint()
        self.hook_k_input = HookPoint()
        self.hook_v_input = HookPoint()
        self.hook_mlp_in = HookPoint()
        self.hook_attn_out = HookPoint()
        self.hook_mlp_out = HookPoint()
        self.hook_resid_pre = HookPoint()

        if not self.cfg.attn_only:
            self.hook_resid_mid = HookPoint()

        self.hook_resid_post = HookPoint()

    def forward(
            self,
            resid_pre: Float[torch.Tensor, "batch pos d_model"],
    ) -> Float[torch.Tensor, "batch pos d_model"]:
        
        resid_pre = self.hook_resid_pre(resid_pre)

        if self.cfg.use_attn_in or self.cfg.use_split_qkv_input:
            # We're adding a head dimension
            attn_in = add_head_dimension(resid_pre, self.cfg.n_heads, clone_tensor=False)
        else:
            attn_in = resid_pre

        if self.cfg.use_attn_in:
            attn_in = self.hook_attn_in(attn_in.clone())
        
        if self.cfg.use_split_qkv_input:
            query_input = self.hook_q_input(attn_in.clone())
            key_input = self.hook_k_input(attn_in.clone())
            value_input = self.hook_v_input(attn_in.clone())
        else:
            query_input = attn_in
            key_input = attn_in
            value_input = attn_in
        
        attn_out = self.attn(
                query_input,
                key_input,
                value_input,
            )

        # Take hook fn
        
        attn_out = self.hook_attn_out(
            attn_out
        )
        
        attn_out = self.ln1(attn_out)

        if not self.cfg.attn_only: 
            resid_mid = self.hook_resid_mid(
                resid_pre + attn_out
            )
            mlp_in = (
                resid_mid
                if not self.cfg.use_hook_mlp_in
                else self.hook_mlp_in(resid_mid.clone())
            )
            normalized_resid_mid = mlp_in
            mlp_out = self.hook_mlp_out(self.mlp(normalized_resid_mid))
            
            mlp_out = self.ln2(mlp_out)
            
            resid_post = self.hook_resid_post(resid_mid + mlp_out)
        else:
            resid_post = self.hook_resid_post(resid_pre + attn_out)
        
        return resid_post



import unittest
import torch
from vit_prisma.models.layers.attention import Attention  
from vit_prisma.models.layers.mlp import MLP  
from vit_prisma.models.layers.transformer_block import TransformerBlock
from vit_prisma.models.base_vit import HookedViT
from vit_prisma.models.layers.patch_embedding import PatchEmbedding

import logging

import torch.nn as nn

from vit_prisma.configs.HookedViTConfig import HookedViTConfig


def get_test_config():
    n_layers = 1
    d_head = 8
    d_model = 8
    d_mlp=8
    return HookedViTConfig(n_layers=n_layers, d_head=d_head, d_model=d_model, d_mlp=d_mlp)

class TestAttention(unittest.TestCase):
    def test_attention(self):
        config = get_test_config()
        attention = Attention(config)
        x = torch.randn(8, 16, config.d_model)

        # Test forward pass
        output = attention(x, x, x)
        self.assertEqual(output.shape, (8, 16, config.d_model))


class TestMLP(unittest.TestCase):
    def test_mlp(self):
        config = get_test_config()
        mlp = MLP(config)
        x = torch.randn(8, 16, config.d_model)

        # Test forward pass
        output = mlp(x)
        self.assertEqual(output.shape, (8, 16, config.d_model))

class TestTransformerBlock(unittest.TestCase):
    def test_transformer_block(self):
        config = get_test_config()
        transformer_block = TransformerBlock(config)
        x = torch.randn(8, 16, config.d_model) 

        # Test forward pass
        output = transformer_block(x)
        self.assertEqual(output.shape, (8, 16, config.d_model))

class TestHookedViT(unittest.TestCase):
    def test_hooked_vit(self):
        config = get_test_config()
        config.return_type = "class_logits"
        model = HookedViT(config)
        
        x = torch.randn(8, config.n_channels, config.image_size, config.image_size)

        # Test forward pass
        output = model(x)
        self.assertEqual(output.shape, (8, config.n_classes))

class TestPatchEmbedding(unittest.TestCase):
    def setUp(self):
        logging.basicConfig(level=logging.INFO)
        self.logger = logging.getLogger(__name__)

    def test_patch_embedding(self):
        config = get_test_config()
        patch_embedding = PatchEmbedding(config, self.logger)
        
        x = torch.randn(8, config.n_channels, config.image_size, config.image_size)


        # Calculate expected number of patches
        num_patches = (config.image_size // config.patch_size) ** 2

        # Test forward pass
        output = patch_embedding(x)
        self.assertEqual(output.shape, (8, num_patches, config.d_model))


if __name__ == '__main__':
    unittest.main()
